"""Tests for delete CLI functions"""

import logging
from click import Context
from click.testing import CliRunner

from housekeeper.cli import delete
from housekeeper.store.api.core import Store
from housekeeper.store.models import Bundle


def test_delete_files_non_specified(base_context: Context, cli_runner: CliRunner, caplog):
    """Test to delete files without specifying bundle name or tag"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a store and a cli runner

    # WHEN trying to delete files without specifying bundle name or tag
    result = cli_runner.invoke(delete.files_cmd, [], obj=base_context)

    # THEN assert it exits non zero
    assert result.exit_code == 1

    # THEN HAL9000 should interfere
    assert "Please specify" in caplog.text


def test_delete_files_non_existing_bundle(
    base_context: Context, cli_runner: CliRunner, case_id: str, caplog
):
    """Test to delete a non existing bundle"""
    caplog.set_level(logging.DEBUG)

    # GIVEN a context with a store and a cli runner

    # WHEN trying to delete a bundle
    result = cli_runner.invoke(delete.files_cmd, ["--bundle-name", case_id], obj=base_context)

    # THEN assert it exits non zero
    assert result.exit_code == 1

    # THEN it should communicate that the bundle was not found
    assert f"Bundle {case_id} not found" in caplog.text


def test_delete_existing_bundle_with_confirmation(
    populated_context: Context, cli_runner: CliRunner, case_id: str, caplog
):
    """Test to delete an existing bundle with confirmation"""
    caplog.set_level(logging.DEBUG)

    # GIVEN a context with a populated store and a cli runner
    store: Store = populated_context["store"]

    # GIVEN a existing bundle
    bundle: Bundle = store._get_query(table=Bundle).first()
    assert bundle
    case_id = bundle.name

    # WHEN trying to delete files without specifying bundle name or tag
    result = cli_runner.invoke(delete.files_cmd, ["--bundle-name", case_id], obj=populated_context)

    # THEN it should ask if you are sure
    assert "Are you sure you want to delete" in result.output


def test_delete_existing_bundle_no_confirmation(
    populated_context: Context, cli_runner: CliRunner, case_id: str, caplog
):
    """Test to delete an existing bundle without confirmation"""
    caplog.set_level(logging.DEBUG)

    # GIVEN a context with a populated store and a cli runner
    store: Store = populated_context["store"]

    # GIVEN a existing bundle
    bundle: Bundle = store._get_query(table=Bundle).first()
    assert bundle

    case_id = bundle.name

    # GIVEN the bundle files
    files = store.get_files_before(bundle_name=case_id, tag_names=[])
    nr_files = len(files)
    assert nr_files > 0

    # WHEN trying to delete a bundle without requiring confirmation
    cli_runner.invoke(delete.files_cmd, ["--bundle-name", case_id, "--yes"], obj=populated_context)

    # THEN the bundle should have been removed
    assert "deleted" in caplog.text



from enum import Enum
from typing import Callable

from sqlalchemy import and_
from sqlalchemy.orm import Query
from sqlalchemy.sql.elements import BooleanClauseList

from housekeeper.store.models import Archive

ARCHIVING_ONGOING: BooleanClauseList = and_(Archive.archiving_task_id, Archive.archived_at == None)
RETRIEVAL_ONGOING: BooleanClauseList = and_(Archive.retrieval_task_id, Archive.retrieved_at == None)


def filter_archiving_ongoing(archives: Query, **kwargs) -> Query:
    """Return archives where the archiving is not marked as completed."""
    return archives.filter(ARCHIVING_ONGOING)


def filter_retrieval_ongoing(archives: Query, **kwargs) -> Query:
    """Return archives where the retrieval is not marked as completed."""
    return archives.filter(RETRIEVAL_ONGOING)


def filter_by_archiving_task_id(archives: Query, task_id: int, **kwargs) -> Query:
    """Return archives where the archiving task id matches the one given."""
    return archives.filter(Archive.archiving_task_id == task_id)


def filter_by_retrieval_task_id(archives: Query, task_id: int, **kwargs) -> Query:
    """Return archives where the retrieval task id matches the one given."""
    return archives.filter(Archive.retrieval_task_id == task_id)


class ArchiveFilter(Enum):
    """Define Archive filter functions."""

    FILTER_ARCHIVING_ONGOING: Callable = filter_archiving_ongoing
    FILTER_RETRIEVAL_ONGOING: Callable = filter_retrieval_ongoing
    FILTER_BY_ARCHIVING_TASK_ID: Callable = filter_by_archiving_task_id
    FILTER_BY_RETRIEVAL_TASK_ID: Callable = filter_by_retrieval_task_id


def apply_archive_filter(
    archives: Query, filter_functions: list[ArchiveFilter], task_id: int = None
) -> Query:
    """Apply filtering functions to archives and return filtered Query."""
    for filter_function in filter_functions:
        archives: Query = filter_function(archives=archives, task_id=task_id)
    return archives



"""Tests for get version CLI functionality"""
import logging

from datetime import datetime
from click import Context
from click.testing import CliRunner

from housekeeper.cli.get import version_cmd
from housekeeper.store.api.core import Store
from housekeeper.store.models import Bundle, Version


def test_get_version_no_input(populated_context: Context, cli_runner: CliRunner, caplog):
    """Test to get all versions when no input is given

    This should exit with False
    """
    caplog.set_level(logging.DEBUG)

    # GIVEN a context that is populated
    # WHEN running the get versions command
    result = cli_runner.invoke(version_cmd, [], obj=populated_context)

    # THEN assert that the program exits with a zero exit code
    assert result.exit_code == 0
    # THEN assert that the correct information is communicated
    assert "Please select a bundle or a version" in caplog.text


def test_get_version_bundle_name(populated_context: Context, cli_runner: CliRunner, helpers):
    """Test to get all versions from a bundle by using the bundle name.

    This should return all versions
    """
    # GIVEN a context that is populated
    store: Store = populated_context["store"]
    bundle_obj: Bundle = store._get_query(table=Bundle).first()
    version_obj: Version = bundle_obj.versions[0]
    bundle_name: str = bundle_obj.name

    # WHEN running the include files command
    result = cli_runner.invoke(version_cmd, ["-b", bundle_name], obj=populated_context)

    # THEN assert that the version path was created
    assert str(version_obj.id) in result.output


def test_get_version_non_existing_bundle_name(
    populated_context: Context, cli_runner: CliRunner, caplog
):
    """
    Test to get all versions from a bundle that does not exist
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN a context that is populated
    store: Store = populated_context["store"]
    # GIVEN a non existing bundle name
    bundle_name = "hello"
    assert not store.get_bundle_by_name(bundle_name=bundle_name)

    # WHEN running the include files command
    cli_runner.invoke(version_cmd, ["-b", bundle_name], obj=populated_context)

    # THEN assert that the correct information is communicated
    assert f"Could not find bundle {bundle_name}" in caplog.text


def test_get_version_json(populated_context: Context, cli_runner: CliRunner, helpers):
    """Test to get all versions from a bundle by using the bundle name in json format.

    This should return all versions
    """
    # GIVEN a context that is populated
    store: Store = populated_context["store"]
    bundle_obj = store._get_query(table=Bundle).first()
    bundle_name = bundle_obj.name

    # WHEN running the include files command
    result = cli_runner.invoke(version_cmd, ["-b", bundle_name, "--json"], obj=populated_context)

    # THEN assert that the program exits succesfully
    assert result.exit_code == 0


def test_get_version_version_id(populated_context: Context, cli_runner: CliRunner, helpers):
    """Test to get all versions from a bundle by using the bundle name in json format.

    This should return all versions
    """
    # GIVEN a context that is populated
    store: Store = populated_context["store"]
    bundle_obj: Bundle = store._get_query(table=Bundle).first()
    version_obj: Version = bundle_obj.versions[0]
    version_id: int = version_obj.id

    # WHEN running the include files command
    result = cli_runner.invoke(version_cmd, ["-i", version_id], obj=populated_context)

    # THEN assert that the program exits succesfully
    assert result.exit_code == 0



"""Constants for housekeeper"""
from datetime import timedelta

LOGLEVELS = ["DEBUG", "INFO", "WARNING", "CRITICAL"]
PIPELINES = ("mip",)
# three months before clean up by default
TIME_TO_CLEANUP = timedelta(days=(30 * 3))
ARCHIVE_TYPES = ("data", "result", "meta", "archive")
EXTRA_STATUSES = ["coverage", "frequency", "genotype", "visualizer", "rawdata", "qc"]
ROOT: str = "root"



"""Fixtures for the CLI add tests"""
import json
from pathlib import Path

import pytest


@pytest.fixture
def file_data(second_sample_vcf: Path, case_id: str, sample_tag_names: list[str]) -> dict:
    """Return a dictionary with file information"""
    return {"path": str(second_sample_vcf), "tags": sample_tag_names, "bundle": case_id}


@pytest.fixture
def file_data_json(file_data) -> str:
    """Return a json string with file information"""
    return json.dumps(file_data)



"""Tests for adding bundles via CLI"""
import json
import logging

from click import Context
from click.testing import CliRunner

from housekeeper.cli.add import bundle_cmd
from housekeeper.store.api.core import Store
from housekeeper.store.models import Bundle


def test_add_existing_bundle(populated_context: Context, cli_runner: CliRunner, caplog):
    """Test to add a bundle that exists"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a populated store and a cli runner
    store: Store = populated_context["store"]
    # GIVEN a existing bundle
    bundle_obj = store._get_query(table=Bundle).first()
    assert bundle_obj
    bundle_name = bundle_obj.name

    # WHEN trying to add a bundle that exists
    result = cli_runner.invoke(bundle_cmd, [bundle_name], obj=populated_context)

    # THEN assert it has a non zero exit status
    assert result.exit_code == 1
    # THEN check that the error message is displayed
    assert "already exists" in caplog.text


def test_add_simple_bundle(base_context: Context, cli_runner: CliRunner, case_id: str, caplog):
    """Test to add a new bundle by only specifying bundle name"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a empty store and a cli runner
    bundle_name = case_id

    # WHEN trying to add a bundle with only the bundle name
    result = cli_runner.invoke(bundle_cmd, [bundle_name], obj=base_context)

    # THEN assert it succeded
    assert result.exit_code == 0
    # THEN check that the proper information is displayed
    assert "new bundle added" in caplog.text


def test_add_bundle_no_input(base_context: Context, cli_runner: CliRunner, caplog):
    """Test to add bundle without any input"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a empty store and a cli runner

    # WHEN running the add bundle function without input
    result = cli_runner.invoke(bundle_cmd, [], obj=base_context)

    # THEN the command failed since there where no input
    assert result.exit_code == 1
    # THEN check that the proper information is displayed
    assert "input json or bundle_name" in caplog.text


def test_add_bundle_json_and_bundle_name(
    base_context: Context,
    cli_runner: CliRunner,
    bundle_data_json: str,
    case_id: str,
    caplog,
):
    """Test to add a new bundle using json AND bundle_name as input.

    This test should fail since it will be confusing to provide inconsistent information
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a empty store, a cli runner and a bundle in json format
    # WHEN trying to add a bundle with both json and bundle_name as input
    result = cli_runner.invoke(bundle_cmd, [case_id, "--json", bundle_data_json], obj=base_context)

    # THEN assert the call failed
    assert result.exit_code == 1
    # THEN check that the proper information is displayed
    assert "not input both json and bundle_name" in caplog.text


def test_add_bundle_json(
    base_context: Context, cli_runner: CliRunner, bundle_data_json: str, caplog
):
    """Test to add a new bundle using json as input"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a empty store, a cli runner and a bundle in json format
    # WHEN trying to add a bundle
    result = cli_runner.invoke(bundle_cmd, ["--json", bundle_data_json], obj=base_context)

    # THEN assert it succeded
    assert result.exit_code == 0
    # THEN check that the proper information is displayed
    assert "new bundle added" in caplog.text


def test_add_bundle_json_no_files(
    base_context: Context, cli_runner: CliRunner, bundle_data_json: str, caplog
):
    """Test to add a new bundle using json as input but no files. It should be possible to add
    a bundle without any files
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a empty store, a cli runner and a bundle in json format without files
    bundle_data = json.loads(bundle_data_json)
    bundle_data.pop("files")
    bundle_data_json = json.dumps(bundle_data)

    # WHEN trying to add a bundle
    result = cli_runner.invoke(bundle_cmd, ["--json", bundle_data_json], obj=base_context)

    # THEN assert it succeded
    assert result.exit_code == 0
    # THEN check that the proper information is displayed even if there where no files added
    assert "new bundle added" in caplog.text


def test_add_bundle_non_existing_file(
    base_context: Context, cli_runner: CliRunner, bundle_data_json: str, caplog
):
    """Test to add a new bundle using json as input when some files does not exist.

    The program should exit with a non zero exit code since a file is missing
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a empty store, a cli runner and a bundle in json format with non
    # existing files
    non_existing = "a_non_existing_file.txt"
    bundle_data = json.loads(bundle_data_json)
    bundle_data["files"][0]["path"] = non_existing
    bundle_data_json = json.dumps(bundle_data)

    # WHEN trying to add a bundle
    result = cli_runner.invoke(bundle_cmd, ["--json", bundle_data_json], obj=base_context)

    # THEN assert it succeded
    assert result.exit_code == 1
    # THEN check that the proper information is displayed
    assert f"File {non_existing} does not exist" in caplog.text


def test_add_bundle_json_missing_data(
    base_context: Context, cli_runner: CliRunner, bundle_data_json: str, caplog
):
    """Test to add a new bundle using json as input but no date. It is mandatory to have a
    created_at date.
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a empty store, a cli runner and a bundle in json format without files
    bundle_data = json.loads(bundle_data_json)
    bundle_data.pop("created_at")
    bundle_data_json = json.dumps(bundle_data)

    # WHEN trying to add a bundle
    result = cli_runner.invoke(bundle_cmd, ["--json", bundle_data_json], obj=base_context)

    # THEN assert it succeded
    assert result.exit_code == 1
    # THEN check that the proper information is displayed even if there where no files added
    assert "Bundle date is required" in caplog.text



"""Base fixtures"""
import copy
import datetime
import json
import shutil
from copy import deepcopy
from pathlib import Path

import pytest
import yaml

from housekeeper.date import get_date
from housekeeper.store import Store
from housekeeper.store.database import create_all_tables, drop_all_tables, initialize_database
from housekeeper.store.models import Archive, Bundle, Tag, Version
from tests.helper_functions import Helpers

# basic fixtures


@pytest.fixture(scope="function")
def helpers() -> Helpers:
    """Return a test helper object."""
    return Helpers()


@pytest.fixture(scope="function")
def vcf_tag_name() -> str:
    """Return a tag named 'vcf'."""
    return "vcf"


@pytest.fixture(scope="function")
def family_tag_name() -> str:
    """Return a tag named 'family'."""
    return "family"


@pytest.fixture(scope="function")
def sample_tag_name() -> str:
    """Return a tag named 'sample'."""
    return "sample"


@pytest.fixture(scope="function")
def non_existent_tag_name() -> str:
    """Return a non-existent tag name."""
    return "not_a_tag_name"


@pytest.fixture(scope="function")
def family_tag_names(vcf_tag_name: str, family_tag_name: str) -> list[str]:
    """Return a list of the family tag names."""
    return [vcf_tag_name, family_tag_name]


@pytest.fixture(scope="function")
def sample_tag_names(vcf_tag_name: str, sample_tag_name: str) -> list[str]:
    """Return a list of the sample tag names."""
    return [vcf_tag_name, sample_tag_name]


@pytest.fixture(scope="session")
def spring_tag() -> str:
    """Return the tag marking SPRING files."""
    return "spring"


@pytest.fixture(scope="session")
def sample_id() -> str:
    """Return name of a sample."""
    return "ACC123456A1"


@pytest.fixture(scope="function")
def archiving_task_id() -> int:
    """Return an id of an archiving task."""
    return 1234


@pytest.fixture(scope="session")
def new_archiving_task_id() -> int:
    """Return a new id of an archiving task."""
    return 1235


@pytest.fixture(scope="function")
def retrieval_task_id() -> int:
    """Return an id of a retrieval task."""
    return 4321


@pytest.fixture(scope="function")
def case_id() -> str:
    """Return name of a case."""
    return "handsomepig"


@pytest.fixture(scope="function")
def other_case_id() -> str:
    """Return name of a another case."""
    return "maturecogar"


@pytest.fixture(scope="function")
def sample_data(sample_tag_names: list[str], sample_vcf: Path) -> dict:
    """Return file and tags for sample."""
    return {"tags": sample_tag_names, "file": sample_vcf}


@pytest.fixture(scope="function")
def sample2_data(sample_tag_names: list[str], second_sample_vcf: Path) -> dict:
    """Return file and tags for sample."""
    return {"tags": sample_tag_names, "file": second_sample_vcf}


@pytest.fixture(scope="function")
def spring_file_1_with_tags(sample_id: str, spring_tag: str, spring_file_1: Path) -> dict:
    """Return spring file and tags for a sample."""
    return {"tags": [sample_id, spring_tag], "file": spring_file_1}


@pytest.fixture(scope="function")
def spring_file_2_with_tags(sample_id: str, spring_tag: str, spring_file_2: Path) -> dict:
    """Return a second SPRING file and tags for a sample."""
    return {"tags": [sample_id, spring_tag], "file": spring_file_2}


@pytest.fixture(scope="session")
def spring_file_3_with_tags(sample_id: str, spring_tag: str, spring_file_3: Path) -> dict:
    """Return a third SPRING file and tags for a sample."""
    return {"tags": [sample_id, spring_tag], "file": spring_file_3}


@pytest.fixture(scope="function")
def family_data(family_tag_names: list[str], family_vcf: Path) -> dict:
    """Return file and tags for sample."""
    return {"tags": family_tag_names, "file": family_vcf}


@pytest.fixture(scope="function")
def family2_data(family_tag_names: list[str], second_family_vcf: Path) -> dict:
    """Return file and tags for sample."""
    return {"tags": family_tag_names, "file": second_family_vcf}


@pytest.fixture(scope="function")
def family3_data(family_tag_names: list[str], third_family_vcf: Path) -> dict:
    """Return file and tags for sample."""
    return {"tags": family_tag_names, "file": third_family_vcf}


@pytest.fixture(scope="function")
def timestamp_string() -> str:
    """Return a time stamp in str format."""
    return "2020-05-01"


@pytest.fixture(scope="function")
def timestamp(timestamp_string: str) -> datetime.datetime:
    """Return a time stamp in date time format."""
    return get_date(timestamp_string)


@pytest.fixture(scope="function")
def later_timestamp() -> datetime.datetime:
    """Return a time stamp in date time format to a later date."""
    return datetime.datetime(2020, 5, 25)


@pytest.fixture(scope="function")
def bundle_data(
    case_id: str,
    sample_data: dict,
    family_data: dict,
    timestamp: datetime.datetime,
    helpers: Helpers,
) -> dict:
    """Return a bundle."""
    data = helpers.create_bundle_data(
        case_id=case_id, files=[family_data, sample_data], created_at=timestamp
    )
    return data


@pytest.fixture(scope="function")
def sample_bundle_data(
    sample_id: str,
    sample_data: dict,
    spring_file_1_with_tags: dict,
    spring_file_2_with_tags: dict,
    spring_file_3_with_tags: dict,
    timestamp: datetime.datetime,
    helpers: Helpers,
) -> dict:
    """Return a bundle containing mock sequencing files."""
    return helpers.create_bundle_data(
        case_id=sample_id,
        files=[spring_file_1_with_tags, spring_file_2_with_tags, spring_file_3_with_tags],
        created_at=timestamp,
    )


@pytest.fixture(scope="function")
def empty_version_data(later_timestamp: datetime.datetime, case_id: str) -> dict:
    """Return a dummy bundle."""
    data = {"bundle_name": case_id, "created_at": later_timestamp, "files": []}
    return data


@pytest.fixture(scope="function")
def version_data(empty_version_data: dict, family2_data: dict, sample2_data: dict) -> dict:
    """Return a dummy bundle."""
    data = copy.deepcopy(empty_version_data)
    data["files"] = [
        {
            "path": str(sample2_data["file"]),
            "archive": False,
            "tags": sample2_data["tags"],
        },
        {
            "path": str(family2_data["file"]),
            "archive": True,
            "tags": family2_data["tags"],
        },
    ]
    return data


@pytest.fixture(scope="function")
def bundle_data_json(bundle_data: dict) -> str:
    """Return a dummy bundle."""
    json_data = copy.deepcopy(bundle_data)
    json_data["created_at"] = str(json_data.pop("created_at"))
    return json.dumps(json_data)


@pytest.fixture(scope="function")
def empty_version_data_json(empty_version_data: dict) -> str:
    """Return a dummy bundle."""
    json_data = copy.deepcopy(empty_version_data)
    json_data["created_at"] = str(json_data.pop("created_at"))
    return json.dumps(json_data)


@pytest.fixture(scope="function")
def version_data_json(version_data: dict) -> str:
    """Return a dummy bundle."""
    json_data = copy.deepcopy(version_data)
    json_data["created_at"] = str(json_data.pop("created_at"))
    return json.dumps(json_data)


@pytest.fixture(scope="function")
def other_bundle(
    bundle_data: dict,
    other_case_id: str,
    later_timestamp: datetime.datetime,
    second_sample_vcf: Path,
    second_family_vcf: Path,
) -> dict:
    """Return a dummy bundle."""
    data = deepcopy(bundle_data)
    data["name"] = other_case_id
    data["created_at"] = later_timestamp
    data["files"][0]["path"] = str(second_sample_vcf)
    data["files"][1]["path"] = str(second_family_vcf)
    return data


@pytest.fixture(scope="function")
def db_name() -> str:
    """Return the name of a database."""
    return "hk_test.db"


@pytest.fixture(scope="function")
def db_path(db_dir: Path, db_name: str) -> Path:
    """Return the path to a database."""
    return db_dir / db_name


@pytest.fixture(scope="function")
def db_uri(db_path: Path) -> str:
    """Return the uri to an in memory database."""
    return "sqlite:///" + str(db_path)


@pytest.fixture(scope="function")
def db_uri_memory() -> str:
    """Return the uri to an in memory database."""
    return "sqlite:///:memory:"


@pytest.fixture(scope="function")
def configs(project_dir: Path, db_uri: str) -> dict:
    """Return a dict with housekeeper configs."""
    _configs = {"root": str(project_dir), "database": db_uri}
    return _configs


# object fixtures


@pytest.fixture(scope="function")
def vcf_tag_obj(vcf_tag_name: str, timestamp: datetime.datetime) -> str:
    """Return a tag object."""
    return Tag(name=vcf_tag_name, created_at=timestamp)


@pytest.fixture(scope="function")
def bundle_obj(bundle_data: dict, store: Store) -> Bundle:
    """Return a bundle object."""
    return store.add_bundle(bundle_data)[0]


@pytest.fixture(scope="function")
def sample_bundle(sample_bundle_data: dict, store: Store) -> Bundle:
    """Return a bundle object."""
    return store.add_bundle(sample_bundle_data)[0]


@pytest.fixture(scope="function")
def version_obj(bundle_data: dict, store: Store) -> Version:
    """Return a version object."""
    return store.add_bundle(bundle_data)[1]


@pytest.fixture(scope="function")
def archive(populated_store: Store) -> Archive:
    """Return an archive object."""
    return populated_store._get_query(table=Archive).first()


# dir fixtures


@pytest.fixture(scope="session")
def fixtures_dir() -> Path:
    """Return the path to the fixtures directory."""
    return Path("tests/fixtures/")


@pytest.fixture(scope="function")
def vcf_dir(fixtures_dir: Path) -> Path:
    """Return the path to the vcf fixtures directory."""
    return fixtures_dir / "vcfs"


@pytest.fixture(scope="session")
def sequencing_files_dir(fixtures_dir: Path) -> Path:
    """Return the path to the sequencing_files fixtures directory."""
    return Path(fixtures_dir, "sequencing_files")


@pytest.fixture(scope="function")
def project_dir(tmpdir_factory) -> Path:
    """Path to a temporary working directory."""
    my_tmpdir = Path(tmpdir_factory.mktemp("workdir"))
    yield my_tmpdir
    shutil.rmtree(str(my_tmpdir))


@pytest.fixture(scope="function")
def housekeeper_version_dir(project_dir: Path, case_id: str, timestamp_string: str) -> Path:
    """Path to a created directory with case and version."""
    hk_version_tmpdir: Path = Path(project_dir, case_id, timestamp_string)
    hk_version_tmpdir.mkdir(parents=True)
    return hk_version_tmpdir


@pytest.fixture(scope="function")
def config_dir(tmpdir_factory) -> Path:
    """Path to a temporary directory for config files."""
    my_tmpdir = Path(tmpdir_factory.mktemp("confdir"))
    yield my_tmpdir
    shutil.rmtree(str(my_tmpdir))


@pytest.fixture(scope="function")
def db_dir(tmpdir_factory) -> Path:
    """Path to a temporary directory for databases."""
    my_tmpdir = Path(tmpdir_factory.mktemp("db_dir"))
    yield my_tmpdir
    shutil.rmtree(str(my_tmpdir))


# File fixtures


@pytest.fixture(scope="function")
def config_file(config_dir: Path, configs: dict) -> Path:
    """Create a config file and return the path to it."""
    conf_path = config_dir / "config.json"
    with open(conf_path, "w") as out_file:
        yaml.dump(configs, out_file)
    return conf_path


@pytest.fixture(scope="function")
def sample_vcf(vcf_dir: Path) -> Path:
    """Return the path to a vcf file."""
    return Path(vcf_dir, "example.vcf")


@pytest.fixture(scope="function")
def spring_file_1(sequencing_files_dir: Path) -> Path:
    """Return the path to a SPRING file."""
    return Path(sequencing_files_dir, "lane1.spring")


@pytest.fixture(scope="function")
def spring_file_2(sequencing_files_dir: Path) -> Path:
    """Return the path to a SPRING file."""
    return Path(sequencing_files_dir, "lane2.spring")


@pytest.fixture(scope="session")
def spring_file_3(sequencing_files_dir: Path) -> Path:
    """Return the path to a SPRING file."""
    return Path(sequencing_files_dir, "lane3.spring")


@pytest.fixture(scope="function")
def archived_file(spring_file_1: Path) -> Path:
    """Return the path to an archived file."""
    return spring_file_1


@pytest.fixture(scope="function")
def non_archived_file(spring_file_2: Path) -> Path:
    """Return the path to a non-archived file."""
    return spring_file_2


@pytest.fixture(scope="function")
def spring_file_2(sequencing_files_dir: Path) -> Path:
    """Return the path to a SPRING file."""
    return Path(sequencing_files_dir, "lane2.spring")


@pytest.fixture(scope="function")
def family_vcf(vcf_dir: Path) -> Path:
    """Return the path to a vcf file."""
    return Path(vcf_dir, "family.vcf")


@pytest.fixture(scope="function")
def second_sample_vcf(vcf_dir: Path) -> Path:
    """Return the path to a vcf file."""
    return Path(vcf_dir, "example.2.vcf")


@pytest.fixture(scope="function")
def second_family_vcf(vcf_dir: Path) -> Path:
    """Return the path to a vcf file."""
    return Path(vcf_dir, "family.2.vcf")


@pytest.fixture(scope="function")
def third_family_vcf(vcf_dir: Path) -> Path:
    """Return the path to a vcf file."""
    return Path(vcf_dir, "family.3.vcf")


@pytest.fixture(scope="function")
def checksum_file(fixtures_dir: Path) -> Path:
    """Return the path to file to test checksum."""
    return Path(fixtures_dir, "26a90105b99c05381328317f913e9509e373b64f.txt")


@pytest.fixture(scope="function")
def checksum(checksum_file: Path) -> Path:
    """Return the checksum for checksum test file."""
    return checksum_file.name.rstrip(".txt")


@pytest.fixture(scope="function")
def helpers() -> Helpers:
    """Return a test helper object."""
    return Helpers()


# Store fixtures


@pytest.fixture(scope="function")
def store(project_dir: Path) -> Store:
    """Return a store setup with all tables."""
    initialize_database("sqlite:///")
    _store = Store(root=str(project_dir))
    create_all_tables()
    yield _store
    drop_all_tables()


@pytest.fixture(scope="function")
def populated_store(
    archiving_task_id: int,
    new_archiving_task_id,
    retrieval_task_id: int,
    bundle_data: dict,
    helpers: Helpers,
    sample_bundle_data: dict,
    spring_file_1: Path,
    spring_file_3: Path,
    store: Store,
) -> Store:
    """Returns a populated store."""
    helpers.add_bundle(store=store, bundle=bundle_data)
    helpers.add_bundle(store=store, bundle=sample_bundle_data)
    helpers.add_archive(
        store=store,
        file_id=store.get_files(file_path=spring_file_1.as_posix()).first().id,
        archiving_task_id=archiving_task_id,
    )
    retrieval_archive: Archive = helpers.add_archive(
        store=store,
        file_id=store.get_files(file_path=spring_file_3.as_posix()).first().id,
        archiving_task_id=new_archiving_task_id,
    )
    retrieval_archive.retrieval_task_id = retrieval_task_id
    store.session.commit()

    return store



"""Fixtures for CLI tests"""
import datetime

import pytest
from click.testing import CliRunner
from housekeeper.store import Store
from housekeeper.store.database import create_all_tables, drop_all_tables, initialize_database
from tests.helper_functions import Helpers


@pytest.fixture(scope="function")
def store(project_dir, db_uri):
    """Override the store fixture to get a controlled db path"""
    initialize_database(db_uri)
    _store = Store(root=str(project_dir))
    create_all_tables()
    yield _store
    drop_all_tables()


@pytest.fixture(scope="function")
def cli_runner():
    """Return a cli runner for testing Click"""
    return CliRunner()


@pytest.fixture(scope="function")
def base_context(db_uri, project_dir, store) -> dict:
    """Return a context with initialized database"""
    return {
        "database": db_uri,
        "root": project_dir,
        "store": store,
    }


@pytest.fixture(scope="function")
def populated_context(db_uri, project_dir, populated_store) -> dict:
    """Return a context with initialized database with some data"""
    return {
        "database": db_uri,
        "root": project_dir,
        "store": populated_store,
    }


@pytest.fixture(scope="function")
def populated_store_subsequent(
    store: Store, bundle_data_subsequent: dict, helpers: Helpers
) -> Store:
    """Returns a populated store"""
    helpers.add_bundle(store, bundle_data_subsequent)
    return store


@pytest.fixture(scope="function")
def populated_context_subsequent(db_uri, project_dir, populated_store_subsequent):
    """Return a context with initialized database with some data"""
    return {
        "database": db_uri,
        "root": project_dir,
        "store": populated_store_subsequent,
    }


@pytest.fixture(scope="function")
def bundle_data_subsequent(
    case_id: str,
    family_data: dict,
    family2_data: dict,
    family3_data: dict,
    timestamp: datetime.datetime,
) -> dict:
    """Return a bundle."""
    return {
        "name": case_id,
        "created_at": timestamp,
        "files": [
            {
                "path": str(family_data["file"]),
                "archive": False,
                "tags": family_data["tags"],
            },
            {
                "path": str(family2_data["file"]),
                "archive": False,
                "tags": family2_data["tags"],
            },
            {
                "path": str(family3_data["file"]),
                "archive": True,
                "tags": family3_data["tags"],
            },
        ],
    }



"""Tests for delete CLI functions"""
import logging
import os

import pytest
from click import Context
from click.testing import CliRunner

from housekeeper.cli import delete
from housekeeper.store.api.core import Store
from housekeeper.store.models import Bundle, File, Tag, Version


def test_delete_non_existing_tag(
    non_existent_tag_name: str,
    base_context: Context,
    cli_runner: CliRunner,
    caplog,
):
    """Test trying to delete a non-existent tag."""
    # GIVEN a tag that does not exist

    # WHEN trying to delete the non-existent tag
    result = cli_runner.invoke(delete.tag_cmd, ["--name", non_existent_tag_name], obj=base_context)

    # THEN assert it exits non zero
    assert result.exit_code == 1
    # THEN it should communicate that the tag was not found
    assert f"Tag {non_existent_tag_name} not found" in caplog.text


def test_delete_existing_tag_with_confirmation(
    populated_context: Context, cli_runner: CliRunner, caplog, family_tag_name: str
):
    """Test deleting an existing tag with confirmation."""
    # GIVEN an existing tag in a populated store
    store: Store = populated_context["store"]
    tag: Tag = store.get_tag(tag_name=family_tag_name)
    assert tag

    # WHEN trying to delete the tag
    result = cli_runner.invoke(delete.tag_cmd, ["--name", family_tag_name], obj=populated_context)

    # THEN the confirmation question should be shown in stdout
    assert f"Delete tag {family_tag_name} with" in result.output


def test_delete_existing_tag_no_confirmation(
    populated_context: Context, cli_runner: CliRunner, caplog, family_tag_name: str
):
    """Test deleting an existing tag without confirmation."""
    # GIVEN an existing tag in a populated store
    caplog.set_level(logging.DEBUG)
    store: Store = populated_context["store"]
    tag: Tag = store.get_tag(tag_name=family_tag_name)
    assert tag

    # WHEN trying to delete the tag
    result = cli_runner.invoke(
        delete.tag_cmd, ["--name", family_tag_name, "--yes"], obj=populated_context
    )

    # THEN the tag should have been deleted
    assert result.exit_code == 0
    assert f"Tag {family_tag_name} deleted" in caplog.text
    assert not store.get_tag(tag_name=family_tag_name)


def test_delete_tag_without_deleting_file(
    base_context: Context,
    populated_store: Store,
    cli_runner: CliRunner,
):
    """Test that deleting a tag does not delete files with the tag."""
    # GIVEN a file with a tag
    file_with_tag: File = populated_store.get_file_by_id(file_id=1)
    assert file_with_tag
    tag: Tag = file_with_tag.tags[0]
    assert tag

    # WHEN deleting the tag
    cli_runner.invoke(delete.tag_cmd, ["--name", tag.name, "--yes"], obj=base_context)

    # THEN the tag is deleted
    assert not populated_store.get_tag(tag_name=tag.name)
    # THEN the file is not deleted
    file_without_tag: File = populated_store.get_files(file_path=file_with_tag.path).first()
    assert file_without_tag
    # THEN the file no longer has the tag
    assert tag not in file_without_tag.tags



"""
housekeeper.__main__

The main entry point for the command line interface.

Invoke as `housekeeper` (if installed)
or `python -m housekeeper` (no install required).
"""
import sys

from housekeeper.cli.core import base

if __name__ == "__main__":
    # exit using whatever exit code the CLI returned
    sys.exit(base())



"""
Schema definitions
"""
import marshmallow as ma


class BaseSchema(ma.Schema):
    """
    Base class for marshmallow schemas
    """

    SKIP_VALUES = set([None])

    @ma.post_dump
    def remove_skip_values(self, data):
        """Remove values that are null in db"""
        return {key: value for key, value in data.items() if value not in self.SKIP_VALUES}


class TagSchema(ma.Schema):
    """
    Tag schema definition
    """

    id = ma.fields.Int()
    name = ma.fields.Str(required=True, error_messages={"required": "Tag name is required."})
    category = ma.fields.Str()
    created_at = ma.fields.DateTime()


class FileSchema(ma.Schema):
    """
    File schema definition
    """

    id = ma.fields.Int()
    path = ma.fields.Str(required=True, error_messages={"required": "File path is required."})
    full_path = ma.fields.Str(
        required=True, error_messages={"required": "Full file path is required."}
    )
    archive = ma.fields.Boolean(default=False)
    tags = ma.fields.List(ma.fields.Nested(TagSchema), required=True)


class InputFileSchema(ma.Schema):
    """
    File input schema definition

    This is used for files that are add via json in the CLI
    """

    path = ma.fields.Str(required=True, error_messages={"required": "File path is required."})
    bundle = ma.fields.Str(required=True, error_messages={"required": "Bundle name is required."})
    archive = ma.fields.Boolean(default=False)
    tags = ma.fields.List(ma.fields.Str, required=False)


class VersionSchema(ma.Schema):
    """
    Bundle schema definition
    """

    id = ma.fields.Int()
    created_at = ma.fields.DateTime(
        required=True, error_messages={"required": "Bundle date is required."}
    )
    expires_at = ma.fields.DateTime()
    included_at = ma.fields.DateTime()
    removed_at = ma.fields.DateTime()
    archived_at = ma.fields.DateTime()
    archived_path = ma.fields.Str()
    archived_checksum = ma.fields.Str()
    bundle_id = ma.fields.Int()
    files = ma.fields.List(ma.fields.Nested(FileSchema), required=True)


class InputVersionSchema(ma.Schema):
    """
    Bundle schema definition
    """

    created_at = ma.fields.Str(
        required=True, error_messages={"required": "Bundle date is required."}
    )
    expires_at = ma.fields.Str()
    included_at = ma.fields.Str()
    removed_at = ma.fields.Str()
    archived_at = ma.fields.Str()
    archived_path = ma.fields.Str()
    archived_checksum = ma.fields.Str()
    bundle_name = ma.fields.Str(required=True)
    files = ma.fields.List(ma.fields.Str(), required=False)


class BundleSchema(ma.Schema):
    """
    Bundle schema definition
    """

    id = ma.fields.Int()
    name = ma.fields.Str(required=True, error_messages={"required": "Bundle name is required."})
    created_at = ma.fields.DateTime(
        required=True, error_messages={"required": "Bundle date is required."}
    )
    files = ma.fields.List(ma.fields.Nested(FileSchema), required=True)
    versions = ma.fields.List(ma.fields.Nested(VersionSchema), required=True)


class InputBundleSchema(ma.Schema):
    """
    Bundle input schema definition

    Used when adding bundle via CLI
    """

    name = ma.fields.Str(required=True, error_messages={"required": "Bundle name is required."})
    created_at = ma.fields.Str(
        required=True, error_messages={"required": "Bundle date is required."}
    )
    expires_at = ma.fields.Str(required=False)
    files = ma.fields.List(ma.fields.Str, required=False)



"""Class with helper functions to use in tests"""

import datetime as dt
import json
from typing import Iterable

from housekeeper.constants import LOGLEVELS
from housekeeper.store import Store
from housekeeper.store.models import Archive


class Helpers:
    """Hold small methods that might be helpful for the tests"""

    @staticmethod
    def count_iterable(iter_obj: Iterable) -> int:
        """Count the length of a iterable"""
        return sum(1 for item in iter_obj)

    @staticmethod
    def get_stdout(output: str) -> str:
        """Strip log messages from a string"""
        stripped = []
        # Strip logging lines
        for line in output.split("\n"):
            if any(log_level in line for log_level in LOGLEVELS):
                continue
            stripped.append(line)
        return "\n".join(stripped)

    @staticmethod
    def get_json(output: str) -> list:
        """Convert a string to json"""
        output = Helpers.get_stdout(output)
        return json.loads(output)

    @staticmethod
    def add_bundle(store: Store, bundle: dict) -> None:
        """Add and commit bundle to housekeeper store"""
        bundle_obj, version = store.add_bundle(bundle)
        store.session.add(bundle_obj)
        store.session.add(version)
        store.session.commit()

    @staticmethod
    def add_archive(store: Store, file_id: int, archiving_task_id: int = 1234) -> Archive:
        """Adds an archive object to the database."""
        new_archive: Archive = store.create_archive(
            file_id=file_id, archiving_task_id=archiving_task_id
        )
        store.session.add(new_archive)
        store.session.commit()
        return new_archive

    @staticmethod
    def create_bundle_data(case_id: str, files: list[dict], created_at: dt.datetime = None) -> dict:
        """Create a new bundle_data dictionary with the given parameters."""

        if created_at is None:
            created_at = dt.datetime.now()

        files = [
            {"path": str(file_data["file"]), "archive": True, "tags": file_data["tags"]}
            for file_data in files
        ]

        data = {
            "name": case_id,
            "created_at": created_at,
            "files": files,
        }

        return data



"""Tests for delete version CLI functions"""
import logging
from click import Context
from click.testing import CliRunner

from housekeeper.cli import delete


def test_delete_non_version_no_input(base_context: Context, cli_runner: CliRunner, caplog):
    """Test to delete a version without any input

    The command should fail since there is no input
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a store and a cli runner

    # WHEN trying to delete a version without specifying anything
    result = cli_runner.invoke(delete.version_cmd, [], obj=base_context)

    # THEN assert it exits non zero
    assert result.exit_code == 1
    # THEN it should communicate that the bundle was not found
    assert "Please select a bundle or a version" in caplog.text



from enum import Enum
from typing import Callable

from sqlalchemy.orm import Query

from housekeeper.store.models import Tag


def filter_tag_by_name(tags: Query, tag_name: str) -> Query:
    """Return tag by bundle name."""
    return tags.filter(Tag.name == tag_name)


class TagFilter(Enum):
    """Define Tag filter functions."""

    FILTER_BY_NAME: Callable = filter_tag_by_name


def apply_tag_filter(
    tags: Query,
    filter_functions: list[Callable],
    tag_name: str | None = None,
) -> Query:
    """Apply filtering functions to tag and return filtered Query."""
    for filter_function in filter_functions:
        tags: Query = filter_function(
            tags=tags,
            tag_name=tag_name,
        )
    return tags



"""Tests for delete CLI functions"""

import logging

from click import Context
from click.testing import CliRunner

from housekeeper.cli import delete
from housekeeper.store.api.core import Store
from housekeeper.store.models import File


def test_delete_non_existing_file(base_context: Context, cli_runner: CliRunner, caplog):
    """Test to delete a non existing file"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a store, a file id and a cli runner
    file_id = 1
    # GIVEN that the file does not exist
    store = base_context["store"]
    file_obj: File = store.get_file_by_id(file_id=file_id)
    assert not file_obj

    # WHEN trying to delete the non existing file
    result = cli_runner.invoke(delete.file_cmd, [str(file_id)], obj=base_context)

    # THEN assert it exits non zero
    assert result.exit_code == 1
    # THEN it should communicate that the file was not found
    assert "file not found" in caplog.text


def test_delete_existing_file_with_confirmation(
    populated_context: Context, cli_runner: CliRunner, caplog
):
    """Test to delete an existing file using confirmation"""
    # GIVEN a context with a populated store, a file id and a cli runner
    store: Store = populated_context["store"]
    file_id = 1
    # GIVEN that the file exists
    file_obj: File = store.get_file_by_id(file_id=file_id)
    assert file_obj

    # WHEN trying to delete the file
    result = cli_runner.invoke(delete.file_cmd, [str(file_id)], obj=populated_context)

    # THEN it should ask if you are sure
    assert "Remove file " in result.output


def test_delete_existing_file_no_confirmation(
    populated_context: Context, cli_runner: CliRunner, caplog
):
    """Test to delete a existing file without confirmation"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a populated store, a file id and a cli runner
    store: Store = populated_context["store"]
    file_id = 1
    # GIVEN that the file exists
    file_obj: File = store.get_file_by_id(file_id=file_id)
    assert file_obj

    # WHEN trying to delete the file
    result = cli_runner.invoke(delete.file_cmd, [str(file_id), "--yes"], obj=populated_context)

    # THEN file delete should be in output
    assert "file deleted" in caplog.text



import datetime as dt
from sqlalchemy.orm import Query

from housekeeper.store.api.core import Store
from housekeeper.store.models import Version
from housekeeper.store.filters.version_bundle_filters import filter_version_by_name_and_bundle_date


def test_filter_version_by_name_and_bundle_date_returns_the_correct_version(populated_store: Store):
    """Test getting version by bundle name and date."""

    # GIVEN a store with a version
    version_bundles_query: Query = populated_store._get_join_version_bundle_query()
    assert isinstance(version_bundles_query, Query)
    assert version_bundles_query.count() > 0

    # GIVEN a bundle name and creation date
    bundle_name: str = version_bundles_query.first().bundle.name
    version_date: dt.datetime = version_bundles_query.first().created_at

    # WHEN filtering the version bundles by the given name and date
    filtered_version: Version = filter_version_by_name_and_bundle_date(
        version_bundles=populated_store._get_join_version_bundle_query(),
        bundle_name=bundle_name,
        version_date=version_date,
    ).first()

    # THEN a version should be returned
    assert isinstance(filtered_version, Version)

    # THEN the version's bundle name should match the given name
    assert filtered_version.bundle.name == bundle_name

    # THEN the version's creation date should match the given date
    assert filtered_version.created_at == version_date


def test_filter_version_by_name_and_bundle_date_with_nonexistent_data_returns_empty_query(
    populated_store: Store, time_stamp_now: dt.datetime
):
    """Test that the function returns an empty result when given a non-existent bundle name or version date."""

    # GIVEN a bundle name and version date that does not exist in the store
    nonexistent_bundle_name = "nonexistent_bundle"

    # WHEN filtering the versions by the nonexistent bundle name and version date
    filtered_version: Version = filter_version_by_name_and_bundle_date(
        version_bundles=populated_store._get_join_version_bundle_query(),
        bundle_name=nonexistent_bundle_name,
        version_date=time_stamp_now,
    )

    # THEN the filtered query should be empty
    assert filtered_version.count() == 0



from datetime import datetime

from housekeeper import store
from housekeeper.store import Store
from housekeeper.store.models import Archive
from tests.helper_functions import Helpers


def test_update_archiving_time_stamp_new(archive: Archive, populated_store: store):
    """Tests updating the archived_at timestamp on a given archive when none is present."""
    # GIVEN an Archive with no archived_at timestamp
    assert not archive.archived_at

    # WHEN updating the timestamp
    populated_store.update_archiving_time_stamp(archive=archive)

    # THEN the timestamp should be updated
    assert archive.archived_at


def test_update_archiving_time_stamp_old(
    archive: Archive, old_timestamp: datetime, populated_store: store
):
    """Tests updating the archived_at timestamp on a given archive when there already is one."""
    # GIVEN an Archive with an old archived_at timestamp
    archive.archived_at = old_timestamp

    # WHEN updating the timestamp
    populated_store.update_archiving_time_stamp(archive=archive)

    # THEN the old timestamp should be kept
    assert archive.archived_at == old_timestamp


def test_update_retrieval_time_stamp_new(archive: Archive, populated_store: store):
    """Tests updating the retrieved_at timestamp on a given archive when none is present."""
    # GIVEN an Archive with no retrieved_at timestamp
    assert not archive.retrieved_at

    # WHEN updating the timestamp
    populated_store.update_retrieval_time_stamp(archive=archive)

    # THEN the timestamp should be updated
    assert archive.retrieved_at


def test_update_retrieval_time_stamp_old(
    archive: Archive, old_timestamp: datetime, populated_store: store
):
    """Tests updating the retrieved_at timestamp on a given archive when there already is one."""
    # GIVEN an Archive with an old retrieved_at timestamp
    archive.retrieved_at = old_timestamp

    # WHEN updating the timestamp
    populated_store.update_retrieval_time_stamp(archive=archive)

    # THEN the old timestamp should be kept
    assert archive.retrieved_at == old_timestamp


def test_update_finished_archival_task(
    archive: Archive, archiving_task_id: int, populated_store: Store
):
    """Tests updating all archives matching the given archiving task id."""
    # GIVEN a store with two archives with the same archiving task id
    second_archive: Archive = Helpers.add_archive(
        archiving_task_id=archiving_task_id, file_id=1, store=populated_store
    )

    # WHEN updating all archives with the given archiving task id
    populated_store.update_finished_archival_task(archiving_task_id=archiving_task_id)

    # THEN both archive should have their archived_at timestamp updated

    assert second_archive.archived_at
    assert archive.archived_at
    assert archive.archived_at.minute == second_archive.archived_at.minute


def test_update_finished_retrieval_task(
    archive: Archive, archiving_task_id, retrieval_task_id: int, populated_store: Store
):
    """Tests updating all archives matching the given retrieval task id."""
    # GIVEN a store with two archives with the same retrieval task id
    second_archive: Archive = Helpers.add_archive(
        archiving_task_id=archiving_task_id, file_id=1, store=populated_store
    )
    archive.retrieval_task_id = retrieval_task_id
    second_archive.retrieval_task_id = retrieval_task_id

    # WHEN updating all archives with the given retrieval task id
    populated_store.update_finished_retrieval_task(retrieval_task_id=retrieval_task_id)

    # THEN both archive should have their retrieved_at timestamp updated
    assert archive.retrieved_at
    assert second_archive.retrieved_at
    assert archive.retrieved_at.minute == second_archive.retrieved_at.minute


def test_update_retrieval_task_id(archive: Archive, retrieval_task_id: int, populated_store: store):
    """Tests updating the retrieved_at timestamp on a given archive when there already is one."""
    # GIVEN an Archive with no retrieval_task_id
    assert not archive.retrieval_task_id

    # WHEN updating the retrieval task id
    populated_store.update_retrieval_task_id(archive=archive, retrieval_task_id=retrieval_task_id)

    # THEN the retrieval task id should be set
    assert archive.retrieval_task_id == retrieval_task_id


def test_update_archiving_task_id(
    archive: Archive, new_archiving_task_id: int, populated_store: store
):
    """Tests updating the archiving_task_od on a given archive when there already is one.
    Necessary for retrieval and rearchiving."""
    # GIVEN an Archive with an old archiving_task_id
    assert archive.archiving_task_id != new_archiving_task_id

    # WHEN updating the archiving_task_id
    populated_store.update_archiving_task_id(
        archive=archive, archiving_task_id=new_archiving_task_id
    )

    # THEN the retrieval task id should be set
    assert archive.archiving_task_id == new_archiving_task_id



"""
This module handles updating entries in the store/database.
"""
import logging
from datetime import datetime

from sqlalchemy.orm import Session

from housekeeper.store.api.handlers.base import BaseHandler
from housekeeper.store.filters.archive_filters import (
    ArchiveFilter,
    apply_archive_filter,
)
from housekeeper.store.models import Archive

LOG = logging.getLogger(__name__)


class UpdateHandler(BaseHandler):
    """Handler for updating entries in the database."""

    def __init__(self, session: Session):
        super().__init__(session=session)

    @staticmethod
    def update_archiving_time_stamp(archive: Archive) -> None:
        """Sets the archived_at timestamp of the given archive to now if no previous timestamp is
        present."""
        if archive.archived_at:
            return
        archive.archived_at = datetime.now()

    @staticmethod
    def update_retrieval_time_stamp(archive: Archive) -> None:
        """Sets the retrieved_at timestamp of the given archive to now if no previous timestamp is
        present."""
        if archive.retrieved_at:
            return
        archive.retrieved_at = datetime.now()

    def update_finished_archival_task(self, archiving_task_id: int) -> None:
        """Sets the archived_at field to now for all archives with the given archiving task id."""
        completed_archives: list[Archive] = apply_archive_filter(
            archives=self._get_query(table=Archive),
            filter_functions=[ArchiveFilter.FILTER_BY_ARCHIVING_TASK_ID],
            task_id=archiving_task_id,
        ).all()
        for archive in completed_archives:
            self.update_archiving_time_stamp(archive=archive)

    def update_finished_retrieval_task(self, retrieval_task_id: int) -> None:
        """Sets the archived_at field to now for all archives with the given archiving task id."""
        completed_archives: list[Archive] = apply_archive_filter(
            archives=self._get_query(table=Archive),
            filter_functions=[ArchiveFilter.FILTER_BY_RETRIEVAL_TASK_ID],
            task_id=retrieval_task_id,
        ).all()
        for archive in completed_archives:
            self.update_retrieval_time_stamp(archive=archive)

    @staticmethod
    def update_retrieval_task_id(archive: Archive, retrieval_task_id: int):
        """Sets the retrieval_task_id in the Archive entry for the provided file."""
        archive.retrieval_task_id: int = retrieval_task_id

    @staticmethod
    def update_archiving_task_id(archive: Archive, archiving_task_id: int):
        """Sets the archiving_task_id in the Archive entry for the provided file."""
        archive.archiving_task_id: int = archiving_task_id



from sqlalchemy.orm import Query

from housekeeper.store import Store
from housekeeper.store.models import Bundle
from housekeeper.store.filters.bundle_filters import (
    filter_bundle_by_name,
    filter_bundle_by_id,
)


def test_get_query_returns_query_object(populated_store: Store):
    result = populated_store._get_query(table=Bundle)
    assert isinstance(result, Query)


def test_filter_bundles_by_id_returns_the_correct_bundle(populated_store: Store):
    """Test getting collaboration by internal_id."""

    # GIVEN a store with a bundle
    bundle: Bundle = populated_store._get_query(table=Bundle).first()
    assert isinstance(bundle, Bundle)
    assert bundle

    bundle_id: int = bundle.id

    # WHEN retrieving the bundle by id
    bundle: Bundle = filter_bundle_by_id(
        bundles=populated_store._get_query(table=Bundle),
        bundle_id=bundle_id,
    ).first()

    # THEN a bundle should be returned
    assert isinstance(bundle, Bundle)

    # THEN the id should match
    assert bundle.id == bundle_id


def test_filter_bundles_by_name_returns_the_correct_bundle(populated_store: Store):
    """Test getting bundle by name."""

    # GIVEN a store with a bundle
    bundle: Bundle = populated_store._get_query(table=Bundle).first()
    assert isinstance(bundle, Bundle)
    assert bundle

    bundle_name: str = bundle.name

    # WHEN retrieving the bundle by name
    bundle: Bundle = filter_bundle_by_name(
        bundles=populated_store._get_query(table=Bundle),
        bundle_name=bundle_name,
    ).first()

    # THEN a bundle should be returned
    assert isinstance(bundle, Bundle)

    # THEN the name should match
    assert bundle.name == bundle_name



"""Module for adding via CLI"""
import datetime as dt
import logging
from logging import Logger
from pathlib import Path
from typing import Generator

import click

from housekeeper.constants import ROOT
from housekeeper.date import get_date
from housekeeper.exc import VersionIncludedError
from housekeeper.files import load_json, validate_input
from housekeeper.include import include_version, link_to_relative_path, relative_path
from housekeeper.store import Store
from housekeeper.store.models import Bundle, Tag, Version

LOG: Logger = logging.getLogger(__name__)


def validate_args(arg: str, json: str, arg_name: str) -> None:
    """Check if input is valid

    One of the arguments has to be specified. Both arguments are not allowed
    """
    if not (arg or json):
        LOG.error("Please input json or %s", arg_name)
        raise click.Abort

    if arg and json:
        LOG.warning("Can not input both json and %s", arg_name)
        raise click.Abort


@click.group()
def add():
    """Add things to the store."""


@add.command("bundle")
@click.argument("bundle_name", required=False)
@click.option("-j", "--json", help="Input json string")
@click.option("-e", "--exclude", help="Use to not include bundle when adding it.", is_flag=True)
@click.pass_context
def bundle_cmd(context: click.Context, bundle_name: str, json: str, exclude: bool):
    """Add a new bundle."""
    LOG.info("Running add bundle")
    store: Store = context.obj["store"]

    validate_args(arg=bundle_name, json=json, arg_name="bundle_name")

    data: dict = {"name": bundle_name, "created_at": str(dt.datetime.now())}
    # This is to preserve the behaviour of adding a bundle without providing all information
    if json:
        data: dict = load_json(json)

    bundle_name = data["name"]
    if store.get_bundle_by_name(bundle_name=bundle_name):
        LOG.warning("bundle name %s already exists", bundle_name)
        raise click.Abort

    validate_input(data, input_type="bundle")
    data["created_at"] = get_date(data.get("created_at"))

    if "files" not in data:
        data["files"] = []

    try:
        new_bundle, new_version = store.add_bundle(data)
    except FileNotFoundError as err:
        LOG.warning("File %s does not exist", err)
        raise click.Abort
    store.session.add(new_bundle)
    new_version.bundle: Bundle = new_bundle
    store.session.add(new_version)
    if not exclude:
        try:
            context_root_dir: str = context.obj[ROOT]
            include_version(global_root=context_root_dir, version_obj=new_version)
        except VersionIncludedError as error:
            LOG.error(error.message)
            raise click.Abort
        new_version.included_at = dt.datetime.now()
    store.session.commit()
    LOG.info("new bundle added: %s (%s)", new_bundle.name, new_bundle.id)


@add.command("file")
@click.option("-t", "--tag", "tags", multiple=True, help="tag to associate the file by")
@click.option("-b", "--bundle-name", help="name of bundle that file should be added to")
@click.option("-j", "--json", help="json formatted input")
@click.option(
    "-kip",
    "--keep-input-path",
    is_flag=True,
    default=False,
    show_default=True,
    help="Flag to use the input path in Housekeeper",
)
@click.argument("path", required=False)
@click.pass_context
def file_cmd(
    context: click.Context,
    tags: list[str],
    bundle_name: str,
    json: str,
    keep_input_path: bool,
    path: str,
):
    """Add a file to the latest version of a bundle."""
    LOG.info("Running add file")
    store: Store = context.obj["store"]
    validate_args(arg=path, json=json, arg_name="path")

    data: dict = {}
    if json:
        data: dict = load_json(json)
        validate_input(data, input_type="file")

    file_path: Path = Path(data.get("path", path))
    if not file_path.exists():
        LOG.warning("File: %s does not exist", file_path)
        raise click.Abort

    bundle_name = data.get("bundle", bundle_name)
    bundle = store.get_bundle_by_name(bundle_name=bundle_name)

    if bundle is None:
        LOG.warning("unknown bundle: %s", bundle_name)
        raise click.Abort

    tags = data.get("tags", tags)
    if not keep_input_path:
        version: Version = bundle.versions[0]
        link_to_relative_path(version=version, file_path=file_path, root_path=context.obj[ROOT])
        file_path: Path = relative_path(version=version, file=file_path)
    new_file = store.add_file(file_path=file_path, bundle=bundle, tags=tags)
    store.session.add(new_file)
    store.session.commit()
    LOG.info("new file added: %s (%s)", new_file.path, new_file.id)


@add.command("version")
@click.option("-j", "--json", help="Input in json format")
@click.option("--created-at", help="Date when created")
@click.argument("bundle_name", required=False)
@click.pass_context
def version_cmd(context: click.Context, bundle_name: str, created_at: str, json: str):
    """Add a new version to a bundle."""
    LOG.info("Running add version")
    store: Store = context.obj["store"]

    validate_args(arg=bundle_name, json=json, arg_name="bundle_name")

    data: dict = {"bundle_name": bundle_name, "created_at": created_at}
    if json:
        data: dict = load_json(json)
        bundle_name = data["bundle_name"]

    data["created_at"] = data.get("created_at") or str(dt.datetime.now())
    validate_input(data, input_type="version")

    bundle = store.get_bundle_by_name(bundle_name=bundle_name)
    if bundle is None:
        LOG.warning("unknown bundle: %s", bundle_name)
        raise click.Abort

    data["created_at"] = get_date(data.get("created_at"))
    new_version = store.add_version(data, bundle)

    if not new_version:
        LOG.warning("Seems like version already exists for the bundle")
        raise click.Abort

    store.session.add(new_version)
    store.session.commit()
    LOG.info("new version (%s) added to bundle %s", new_version.id, bundle.name)


@add.command("tag")
@click.argument("tags", nargs=-1)
@click.option("-f", "--file-id", type=int)
@click.pass_context
def tag_cmd(context: click.Context, tags: list[str], file_id: int):
    """Add tags to housekeeper. Use `--file-id` to add tags to existing file"""
    LOG.info("Running add tag")
    store: Store = context.obj["store"]
    file = None
    if not tags:
        LOG.warning("No tags provided")
        raise click.Abort

    if file_id:
        file = store.get_file_by_id(file_id=file_id)

        if not file:
            LOG.warning(f"unable to find file with id: {file_id}")
            raise click.Abort

    tag_name: str
    for tag_name in tags:
        tag: Tag = store.get_tag(tag_name)

        if not tag:
            LOG.info("%s: tag created", tag_name)
            tag: Tag = store.new_tag(tag_name)
            store.session.add(tag)
            store.session.commit()

        if not file:
            continue

        if tag in file.tags:
            LOG.info("%s: tag already added", tag_name)
            continue

        file.tags.append(tag)

    store.session.commit()

    if not file:
        return

    all_tags: Generator = (tag.name for tag in file.tags)
    LOG.info("file tags: %s", ", ".join(all_tags))



# -*- coding: utf-8 -*-
"""This is a core module for the Store API"""

import logging
from pathlib import Path

from housekeeper.store.database import get_session
from housekeeper.store.api.handlers.update import UpdateHandler
from housekeeper.store.api.handlers.create import CreateHandler
from housekeeper.store.api.handlers.read import ReadHandler
from housekeeper.store.models import File, Version

LOG = logging.getLogger(__name__)


class CoreHandler(CreateHandler, ReadHandler, UpdateHandler):
    """Aggregating class for the store api handlers"""

    def __init__(self, session):
        ReadHandler(session)
        CreateHandler(session)
        UpdateHandler(session)


class Store(CoreHandler):

    """
    Handles interactions with the database in the context when a temporary
    database connection is needed, e.g. a command line interface.

    Args:
        uri (str): SQLAlchemy database connection str
    """

    def __init__(self, root: str):
        self.session = get_session()

        LOG.debug("Initializing Store")
        File.app_root = Path(root)
        Version.app_root = Path(root)

        super().__init__(self.session)



from enum import Enum
from typing import Callable

from sqlalchemy import func as sqlalchemy_func
from sqlalchemy.orm import Query

from housekeeper.store.models import File, Tag


def filter_files_by_id(files: Query, file_id: int, **kwargs) -> Query:
    """Filter files by file id."""
    return files.filter(File.id == file_id)


def filter_files_by_path(files: Query, file_path: str, **kwargs) -> Query:
    """Filter files by path."""
    return files.filter(File.path == file_path)


def filter_files_by_tags(files: Query, tag_names: list[str], **kwargs) -> Query:
    """Filter files by tag names."""
    return (
        files.filter(Tag.name.in_(tag_names))
        .group_by(File.id)
        .having(sqlalchemy_func.count(Tag.name) == len(tag_names))
    )


def filter_files_by_is_archived(files: Query, is_archived: bool, **kwargs) -> Query:
    """Filters the query depending on if the files are archived or not."""
    return files.filter((File.archive != None) == is_archived)


class FileFilter(Enum):
    """Define filter functions for Files joined tables."""

    FILTER_BY_ID: Callable = filter_files_by_id
    FILTER_BY_PATH: Callable = filter_files_by_path
    FILTER_FILES_BY_TAGS: Callable = filter_files_by_tags
    FILTER_FILES_BY_IS_ARCHIVED: Callable = filter_files_by_is_archived


def apply_file_filter(
    files: Query,
    filter_functions: list[Callable],
    file_id: int | None = None,
    file_path: str | None = None,
    is_archived: bool | None = None,
    tag_names: list[str] | None = None,
) -> Query:
    """Apply filtering functions and return filtered query."""
    for filter_function in filter_functions:
        files: Query = filter_function(
            files=files,
            file_id=file_id,
            file_path=file_path,
            is_archived=is_archived,
            tag_names=tag_names,
        )
    return files



"""Tests for cli get file functionality"""
from pathlib import Path
from housekeeper.cli.get import files_cmd
from housekeeper.cli.tables import squash_names, _get_suffix
from housekeeper.store.api.core import Store


def test_get_files_no_files(base_context, cli_runner, helpers):
    """Test to get all files when there are no files"""
    # GIVEN a context with a populated store and a cli runner
    store = base_context["store"]
    # GIVEN a store without files
    assert helpers.count_iterable(store.get_files()) == 0

    # WHEN fetching all files by not specifying any file
    json_bundles = helpers.get_json(
        cli_runner.invoke(files_cmd, ["--json"], obj=base_context).output
    )

    # THEN assert that we get a list
    assert isinstance(json_bundles, list)
    # THEN assert that no files where fetched
    assert len(json_bundles) == 0


def test_get_files_json(populated_context, cli_runner, helpers):
    """Test to get all files from a populated store"""
    # GIVEN a context with a populated store and a cli runner
    store: Store = populated_context["store"]
    # GIVEN a store with some files
    nr_files = helpers.count_iterable(store.get_files())
    assert nr_files > 0

    # WHEN fetching all files by not specifying any file
    json_bundles = helpers.get_json(
        cli_runner.invoke(files_cmd, ["--json"], obj=populated_context).output
    )

    # THEN assert that all files where fetched
    assert len(json_bundles) == nr_files


def test_get_files(populated_context, cli_runner):
    """Test to get all files from a populated store in human friendly format"""
    # GIVEN a context with a populated store and a cli runner
    # GIVEN a store with some files
    store: Store = populated_context["store"]
    # GIVEN a file name
    file_obj = store.get_files().first()
    file_name = Path(file_obj.path).name

    # WHEN fetching all files by not specifying any file
    result = cli_runner.invoke(files_cmd, [], obj=populated_context)

    # THEN assert that the file name of one file is displayed
    assert file_name in result.output
    # THEN assert that the full file path is not shown
    assert file_obj.path not in result.output


def test_get_files_tag(populated_context, cli_runner, helpers, vcf_tag_name):
    """Test to get files with a specific tag from a populated store"""
    # GIVEN a context with a populated store and a cli runner
    store: Store = populated_context["store"]
    # GIVEN a store with some files that are tagged
    nr_files = helpers.count_iterable(store.get_files(tag_names=[vcf_tag_name]))
    assert nr_files > 0

    # WHEN fetching all files by not specifying any file
    json_bundles = helpers.get_json(
        cli_runner.invoke(
            files_cmd, ["--tag", vcf_tag_name, "--json"], obj=populated_context
        ).output
    )

    # THEN assert that all files where fetched
    assert len(json_bundles) == nr_files


def test_get_files_multiple_tags(
    populated_context, cli_runner, helpers, vcf_tag_name, family_tag_name
):
    """Test to get files with multiple tags tag from a populated store"""
    # GIVEN a context with a populated store and a cli runner
    store: Store = populated_context["store"]
    # GIVEN a store with some files that are tagged
    nr_files = helpers.count_iterable(store.get_files(tag_names=[vcf_tag_name, family_tag_name]))
    assert nr_files > 0

    # WHEN fetching all files by not specifying any file
    json_bundles = helpers.get_json(
        cli_runner.invoke(
            files_cmd,
            ["--tag", vcf_tag_name, "--tag", family_tag_name, "--json"],
            obj=populated_context,
        ).output
    )

    # THEN assert that all files where fetched
    assert len(json_bundles) == nr_files


def test_get_files_rare_tag(populated_context, cli_runner, helpers, family_tag_name):
    """Test to get files with a tag that is not on all files from a populated store"""
    # GIVEN a context with a populated store and a cli runner
    store: Store = populated_context["store"]
    # GIVEN a store with some files that are tagged
    total_nr_files = helpers.count_iterable(store.get_files())
    nr_tag_files = helpers.count_iterable(store.get_files(tag_names=[family_tag_name]))
    # GIVEN a tag that only fetches a subset of the files
    assert nr_tag_files < total_nr_files

    # WHEN fetching all files with that tag
    json_bundles = helpers.get_json(
        cli_runner.invoke(
            files_cmd, ["--tag", family_tag_name, "--json"], obj=populated_context
        ).output
    )

    # THEN assert that all files where fetched
    assert len(json_bundles) == nr_tag_files


def test_get_files_compact(populated_context_subsequent, cli_runner, family_tag_name, helpers):
    """Test to get all files from a populated store in human friendly format, subsequent names concatenated"""
    # GIVEN an example result file list
    file_list = [
        {"path": "family.vcf", "full_path": "tests/family.vcf", "tags": [], "id": 7},
        {
            "path": "family.2.vcf",
            "full_path": "/tests/family.2.vcf",
            "tags": [],
            "id": 8,
        },
        {
            "path": "family.3.vcf",
            "full_path": "/tests/family.3.vcf",
            "tags": [],
            "id": 9,
        },
    ]

    # WHEN calling `squash_names` on list of files
    squashed = squash_names(file_list)

    # THEN assert that the file names displayed are squashed
    assert len(squashed) < len(file_list)


def test_get_suffix():
    # GIVEN
    dont_split_name1 = "asdf2.asdf.vcf"
    dont_split_name2 = "1123"
    dont_split_name3 = "asdf.vcf"
    dont_split_name4 = "asdf_2.asdf.vcf"

    split_name1 = "asdf1.png"
    split_name2 = "asdf8A8_7777_asdf_8.png"

    # THEN assert filename parsing *does not* split name into (prefix, integer, suffix)
    assert (dont_split_name1, "", "") == _get_suffix(dont_split_name1)
    assert (dont_split_name2, "", "") == _get_suffix(dont_split_name2)
    assert (dont_split_name3, "", "") == _get_suffix(dont_split_name3)
    assert (dont_split_name4, "", "") == _get_suffix(dont_split_name4)

    # THEN assert filename parsing *does* split name into (prefix, integer, suffix)
    assert (split_name1, "", "") != _get_suffix(split_name1)
    assert (split_name2, "", "") != _get_suffix(split_name2)



import datetime as dt
from enum import Enum
from typing import Callable

from sqlalchemy.orm import Query

from housekeeper.store.models import Bundle, Version


def filter_version_by_name_and_bundle_date(
    version_bundles: Query, bundle_name: str, version_date: dt.datetime, **kwargs
) -> Query:
    """Return version by date and bundle name."""
    return version_bundles.filter(Bundle.name == bundle_name, Version.created_at == version_date)


class VersionBundleFilters(Enum):
    """Define Version filter functions."""

    FILTER_BY_DATE_AND_NAME: Callable = filter_version_by_name_and_bundle_date


def apply_version_bundle_filter(
    version_bundles: Query,
    filter_functions: list[Callable],
    bundle_name: str | None = None,
    version_date: dt.datetime | None = None,
) -> Query:
    """Apply filtering functions and return filtered query."""
    for filter_function in filter_functions:
        version_bundles: Query = filter_function(
            version_bundles=version_bundles,
            bundle_name=bundle_name,
            version_date=version_date,
        )
    return version_bundles



"""Module for base CLI"""
import logging

import click
import coloredlogs
import yaml

import housekeeper
from housekeeper.constants import ROOT
from housekeeper.store import Store
from housekeeper.store.database import initialize_database

from . import add, delete, get, include, init

LOG = logging.getLogger(__name__)


@click.group()
@click.option("-c", "--config", type=click.File())
@click.option("-d", "--database", help="path/URI of the SQL database")
@click.option("-r", "--root", type=click.Path(exists=True), help="Housekeeper root dir")
@click.option("-l", "--log-level", default="INFO")
@click.version_option(housekeeper.__version__, prog_name=housekeeper.__title__)
@click.pass_context
def base(
    context: click.Context,
    config: click.File,
    database: str | None,
    root: str | None,
    log_level: str,
):
    """Housekeeper - Access your files!"""
    coloredlogs.install(level=log_level)
    config_values = {}
    if config:
        config_values = yaml.full_load(config)
    context.obj = config_values
    db_path = database or context.obj.get("database")
    root_path = context.obj[ROOT] = root or context.obj.get(ROOT)
    if not db_path:
        LOG.error("Please point to a database")
        raise click.Abort
    if not root_path:
        LOG.error("Please specify a root dir")
        raise click.Abort
    context.obj["database"] = db_path
    LOG.info("Use root path %s", root_path)
    initialize_database(db_path)
    context.obj["store"] = Store(root=root_path)


base.add_command(init.init)
base.add_command(add.add)
base.add_command(include.include)
base.add_command(get.get)
base.add_command(delete.delete)



"""Tests for adding versions via CLI"""
import json
import logging

from click import Context
from click.testing import CliRunner

from housekeeper.cli.add import version_cmd
from housekeeper.store.api.core import Store
from housekeeper.store.models import Bundle


def test_add_version_non_input(populated_context: Context, cli_runner: CliRunner, caplog):
    """Test to add a version to a bundle without providing input

    The CLI should exit with non zero since nothing is defined
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a populated store and a cli runner

    # WHEN trying to add a version to a non existing bundle
    result = cli_runner.invoke(version_cmd, [], obj=populated_context)

    # THEN assert it has a non zero exit status
    assert result.exit_code == 1
    # THEN check that the error message is displayed
    assert "Please input json or bundle_name" in caplog.text


def test_add_version_non_existing_bundle(populated_context: Context, cli_runner: CliRunner, caplog):
    """Test to add a version to a non existing bundle

    The CLI should exit with non zero since the bundle does not exist
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a populated store and a cli runner
    store: Store = populated_context["store"]
    # GIVEN a non existing bundle
    bundle_name = "non_existing"
    bundle: Bundle = store.get_bundle_by_name(bundle_name=bundle_name)
    assert not bundle

    # WHEN trying to add a version to a non existing bundle
    result = cli_runner.invoke(version_cmd, [bundle_name], obj=populated_context)

    # THEN assert it has a non zero exit status
    assert result.exit_code == 1
    # THEN check that the error message is displayed
    assert f"unknown bundle: {bundle_name}" in caplog.text


def test_add_version_existing_bundle(populated_context: Context, cli_runner: CliRunner, caplog):
    """Test to add a version to a existing bundle

    The functionality should work as expected since the bundle exists
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a populated store and a cli runner
    store: Store = populated_context["store"]
    # GIVEN a existing bundle
    bundle: Bundle = store._get_query(table=Bundle).first()
    assert isinstance(bundle, Bundle)

    # GIVEN the name of a existing bundle
    bundle_name: str = bundle.name

    # WHEN trying to add a version to an existing bundle
    result = cli_runner.invoke(version_cmd, [bundle_name], obj=populated_context)

    # THEN assert it has a zero exit status
    assert result.exit_code == 0
    # THEN check that the error message is displayed
    assert f"added to bundle {bundle_name}" in caplog.text
    assert "new version" in caplog.text


def test_add_version_existing_bundle_same_date(
    populated_context: Context, cli_runner: CliRunner, timestamp_string: str, caplog
):
    """Test to add a version to an existing bundle when the date is the same

    It should not be possible to add two versions with the same dates
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a populated store and a cli runner
    store: Store = populated_context["store"]
    # GIVEN a existing bundle
    bundle: Bundle = store._get_query(table=Bundle).first()
    assert isinstance(bundle, Bundle)
    bundle_name: str = bundle.name

    # WHEN trying to add a version to an existing bundle
    result = cli_runner.invoke(
        version_cmd,
        [bundle_name, "--created-at", timestamp_string],
        obj=populated_context,
    )

    # THEN assert it has a non zero exit status
    assert result.exit_code == 1
    # THEN check that the error message is displayed
    assert "version of bundle already added" in caplog.text


def test_add_version_no_files_json(
    populated_context: Context,
    cli_runner: CliRunner,
    empty_version_data_json: str,
    caplog,
):
    """Test to add a new empty version to existing bundle using json as input

    This should work since there is no problem to add a version without files
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a populated store a cli runner
    store: Store = populated_context["store"]
    version_data = json.loads(empty_version_data_json)
    # GIVEN a bundle with one version
    bundle = store.get_bundle_by_name(bundle_name=version_data["bundle_name"])
    assert isinstance(bundle, Bundle)
    assert len(bundle.versions) == 1
    # GIVEN version information without files, in json format
    assert version_data["files"] == []

    # WHEN trying to add the version
    result = cli_runner.invoke(
        version_cmd, ["--json", empty_version_data_json], obj=populated_context
    )

    # THEN assert it succeded
    assert result.exit_code == 0
    # THEN check that the error message is displayed
    assert f"added to bundle {version_data['bundle_name']}" in caplog.text
    assert "new version" in caplog.text
    # THEN assert that the version was added
    bundle = store.get_bundle_by_name(bundle_name=version_data["bundle_name"])
    assert len(bundle.versions) == 2


def test_add_version_with_files_json(
    populated_context: Context, cli_runner: CliRunner, version_data_json: str, caplog
):
    """Test to add a new version with files to existing bundle using json as input"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a populated store a cli runner
    store: Store = populated_context["store"]
    version_data = json.loads(version_data_json)
    # GIVEN a bundle with one version
    bundle = store.get_bundle_by_name(bundle_name=version_data["bundle_name"])
    assert isinstance(bundle, Bundle)
    assert len(bundle.versions) == 1
    # GIVEN version information without files, in json format
    assert version_data["files"] != []

    # WHEN trying to add the version
    result = cli_runner.invoke(version_cmd, ["--json", version_data_json], obj=populated_context)

    # THEN assert it succeded
    assert result.exit_code == 0
    # THEN check that the error message is displayed
    assert f"added to bundle {version_data['bundle_name']}" in caplog.text
    assert "new version" in caplog.text
    # THEN assert that the files where added to the version was added
    bundle = store.get_bundle_by_name(bundle_name=version_data["bundle_name"])
    for version_obj in bundle.versions:
        assert len(version_obj.files) == 2


def test_add_version_with_json_and_bundle_name(
    populated_context, cli_runner, version_data_json: str, caplog
):
    """Test to add a new version to existing bundle using both json and bundle_name as input

    The CLI should fail since it is ambiguous to specify the bundle in two ways
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a populated store a cli runner and a dummy bundle name
    bundle_name = "hello"

    # WHEN trying to add the version
    result = cli_runner.invoke(
        version_cmd, [bundle_name, "--json", version_data_json], obj=populated_context
    )

    # THEN assert it succeded
    assert result.exit_code == 1
    # THEN check that the error message is displayed
    assert "Can not input both json and bundle_name" in caplog.text



from sqlalchemy.orm import Query

from housekeeper.store import Store
from housekeeper.store.filters.tag_filters import (
    TagFilter,
    apply_tag_filter,
    filter_tag_by_name,
)
from housekeeper.store.models import Tag


def test_filter_tag_by_name_returns_correct_tag(populated_store: Store, sample_tag_name: str):
    """Test that fetching a tag by name returns a tag with the expected name."""

    # GIVEN a store with the desired tag
    all_tags: Query = populated_store._get_query(table=Tag)
    all_tag_names: list[str] = [t.name for t in all_tags.all()]
    assert sample_tag_name in all_tag_names

    # WHEN retrieving the tag by name
    tag_query: Tag = filter_tag_by_name(
        tags=all_tags,
        tag_name=sample_tag_name,
    )
    tag: Tag = tag_query.first()

    # THEN the returned object is a Query
    assert isinstance(tag_query, Query)

    # THEN the object in the query is a Tag
    assert isinstance(tag, Tag)

    # THEN the tag name should match the input tag name
    assert tag.name == sample_tag_name


def test_filter_tag_by_name_non_existent_tag(populated_store: Store, non_existent_tag_name):
    """Test that using a non-existent tag name returns an empty query."""
    # GIVEN a tag name not included in a populated store
    all_tags: Query = populated_store._get_query(table=Tag)
    all_tag_names: list[str] = [t.name for t in all_tags.all()]
    assert non_existent_tag_name not in all_tag_names

    # WHEN retrieving the tag by name
    tag_query: Query = filter_tag_by_name(
        tags=all_tags,
        tag_name=non_existent_tag_name,
    )
    assert isinstance(tag_query, Query)

    # THEN the retrieved query is empty
    assert tag_query.count() == 0


def test_filter_tag_by_name_with_none_tag_name(populated_store: Store):
    """Test that using None as tag name argument returns an empty query."""
    # GIVEN a populated store

    # WHEN trying to retrieve a tag with None as name
    tag_query: Query = filter_tag_by_name(
        tags=populated_store._get_query(table=Tag),
        tag_name=None,
    )
    assert isinstance(tag_query, Query)

    # THEN the retrieved query is empty
    assert tag_query.count() == 0


def test_apply_tag_filter_without_tag_name(populated_store: Store):
    """Test that omitting the tag name argument returns an empty query"""
    # GIVEN a populated store

    # WHEN trying to retrieve a tag with None as name
    tag_query: Query = apply_tag_filter(
        tags=populated_store._get_query(table=Tag),
        filter_functions=[TagFilter.FILTER_BY_NAME],
    )
    assert isinstance(tag_query, Query)

    # THEN the retrieved query is empty
    assert tag_query.count() == 0



from sqlalchemy.orm import Query

from housekeeper.store import Store
from housekeeper.store.filters.archive_filters import (
    filter_archiving_ongoing,
    filter_retrieval_ongoing,
    filter_by_archiving_task_id,
    filter_by_retrieval_task_id,
)
from housekeeper.store.models import Archive


def test_filter_archiving_ongoing(populated_store: Store, archiving_task_id: int):
    """Tests filtering of Archives on if their archiving task is finished."""
    # GIVEN a populated store with, an ongoing task

    # WHEN filtering a query by ongoing archiving
    ongoing_archives: Query = filter_archiving_ongoing(
        archives=populated_store._get_query(table=Archive)
    )

    # THEN the archive with the ongoing archiving task should be returned
    assert ongoing_archives.count() > 0
    for archive in ongoing_archives:
        assert archive.archiving_task_id
        assert not archive.archived_at


def test_filter_retrieval_ongoing(archive: Archive, populated_store: Store, retrieval_task_id: int):
    """Tests filtering of Archives on if their retrieval task is finished."""
    # GIVEN a populated store with and an ongoing task
    archive.retrieved_at = None
    archive.retrieval_task_id = retrieval_task_id

    # WHEN filtering a query by ongoing retrieval
    ongoing_archives: Query = filter_retrieval_ongoing(
        archives=populated_store._get_query(table=Archive)
    )
    # THEN the archive with the ongoing retrieval task should be returned
    assert ongoing_archives.count() > 0
    for archive in ongoing_archives:
        assert archive.retrieval_task_id
        assert not archive.retrieved_at


def test_filter_by_archiving_task_id(archiving_task_id: int, populated_store: Store):
    """Tests the filtering of archives by archiving task id."""
    # GIVEN a store with an archive with the given archiving task id

    # WHEN filtering by archiving task id
    archives_with_given_task_id: Query = filter_by_archiving_task_id(
        archives=populated_store._get_query(table=Archive), task_id=archiving_task_id
    )

    # THEN only one with matching archiving task id should be returned
    assert archives_with_given_task_id.count() > 0
    for archive in archives_with_given_task_id:
        assert archive.archiving_task_id == archiving_task_id


def test_filter_by_retrival_task_id(
    archive: Archive, retrieval_task_id: int, populated_store: Store
):
    """Tests the filtering of archives by retrieval task id."""
    # GIVEN a store with an archive with the given retrieval task id
    archive.retrieval_task_id = retrieval_task_id

    # WHEN filtering by retrieval task id
    archives_with_given_task_id: Query = filter_by_retrieval_task_id(
        archives=populated_store._get_query(table=Archive), task_id=retrieval_task_id
    )

    # THEN only one with matching retrieval task id should be returned
    assert archives_with_given_task_id.count() > 0
    for archive in archives_with_given_task_id:
        assert archive.retrieval_task_id == retrieval_task_id



"""Tests for the models"""

from housekeeper.store import models


def test_instantiate_bundle_obj(case_id, timestamp):
    """Test instantiate a Bundle object"""
    # GIVEN a case id and a time stamp
    # WHEN instantiating a bundle
    bundle_obj = models.Bundle(name=case_id, created_at=timestamp)
    # THEN assert it was instantiated as expected
    assert bundle_obj.name == case_id
    assert bundle_obj.created_at == timestamp


def test_instantiate_bundle_obj_no_name():
    """Test instantiate a Bundle object without a name"""
    # WHEN instantiating a bundle without a name
    bundle_obj = models.Bundle()
    # THEN assert it was instantiated as expected
    assert bundle_obj.name is None
    assert bundle_obj.versions == []


def test_version_obj_full_path(project_dir, minimal_bundle_obj):
    """Test the to get the full path from a version object"""
    # GIVEN a bundle object
    timestamp = minimal_bundle_obj.created_at
    # WHEN instantiating the version obj
    version_obj = models.Version(created_at=timestamp, bundle=minimal_bundle_obj)
    # THEN it should point to the correct folder
    root_path = version_obj.full_path
    assert root_path == project_dir / minimal_bundle_obj.name / str(timestamp.date())



# -*- coding: utf-8 -*-


class HousekeeperError(Exception):
    def __init__(self, message):
        self.message = message


class VersionIncludedError(HousekeeperError):
    pass


class BundleValidationError(HousekeeperError):
    pass



"""Tests for store core functions."""
from pathlib import Path

import pytest
from sqlalchemy.exc import IntegrityError

from housekeeper.store.api import schema
from housekeeper.store.api.core import Store
from housekeeper.store.models import Archive, Bundle, File, Tag, Version


def test_schema_with_invalid_input(bundle_data_json):
    """Tests that errors are thrown when validating incorrect input data."""
    # GIVEN input data with missing name of the bundle
    del bundle_data_json["name"]
    # WHEN validating it against the schema
    errors = schema.BundleSchema().validate(bundle_data_json)
    # THEN it should report errors for the field
    assert len(errors) > 0
    assert "name" in errors


# tag tests


def test_create_tag_obj(store: Store, vcf_tag_name):
    """Test to create a tag object."""
    # GIVEN a store and a tag name
    # WHEN creating a tag
    new_tag = store.new_tag(vcf_tag_name)
    # THEN assert a tag object was created
    assert isinstance(new_tag, Tag)
    assert new_tag.name == vcf_tag_name


def test_add_tag(store: Store, vcf_tag_obj):
    """Test to add a tag."""
    # GIVEN a store without tags
    assert store._get_query(table=Tag).count() == 0
    # WHEN adding a tag
    store.session.add(vcf_tag_obj)
    store.session.commit()
    # THEN assert the new tag was added
    assert store._get_query(table=Tag).count() == 1
    assert store._get_query(table=Tag).first() == vcf_tag_obj


# version tests


def test_create_version_obj(store: Store, timestamp):
    """Test to create a version object."""
    # GIVEN a store and a time stamp
    # WHEN creating a version
    new_version = store.new_version(created_at=timestamp, expires_at=timestamp)
    # THEN assert a version object was created
    assert isinstance(new_version, Version)
    assert new_version.created_at == timestamp


# bundle tests


def test_create_bundle_obj(store: Store, bundle_data):
    """Test to create a bundle object."""
    # GIVEN some bundle information
    # WHEN adding the new bundle
    bundle_obj = store.add_bundle(bundle_data)[0]
    # THEN the bundle should have correct name
    assert bundle_obj.name == bundle_data["name"]
    # THEN the bundle should have correct creation time
    assert bundle_obj.created_at == bundle_data["created_at"]
    # THEN assert that a version was added
    assert len(bundle_obj.versions) == 1
    # THEN assert that all files where added to the bundle
    assert len(bundle_obj.versions[0].files) == len(bundle_data["files"])


def test_add_bundle(store: Store, bundle_obj):
    """Test to add a bundle to the store."""
    # GIVEN a store without files, tags, versions or bundles
    assert store._get_query(table=Bundle).count() == 0
    assert store._get_query(table=Tag).count() == 0
    assert store._get_query(table=Version).count() == 0
    assert store._get_query(table=File).count() == 0
    # WHEN adding the new bundle
    store.session.add(bundle_obj)
    store.session.commit()
    # THEN assert that a bundle was added
    assert store._get_query(table=Bundle).count() == 1
    # THEN assert that the bundle is correct
    assert store._get_query(table=Bundle).first() == bundle_obj
    # THEN assert some tags where added
    assert store._get_query(table=Tag).count() > 0
    # THEN assert some files where added
    assert store._get_query(table=File).count() > 0
    # THEN assert we have a version
    assert store._get_query(table=Version).count() > 0


def test_add_bundle_twice(populated_store: Store, bundle_data):
    """Test to add a bundle twice."""
    store = populated_store
    # GIVEN a store ppopulated with a bundle
    assert store._get_query(table=Bundle).count() > 0
    # WHEN adding the same bundle again
    new_bundle = store.add_bundle(bundle_data)
    # THEN it should return None
    assert new_bundle is None


def test_add_two_versions_of_bundle(populated_store: Store, second_bundle_data: dict):
    """Test to add two versions of the same bundle."""
    store: Store = populated_store
    # GIVEN a populated store and some modified bundle data
    starting_bundle_count: int = store._get_query(table=Bundle).count()
    starting_version_count: int = store._get_query(table=Version).count()
    starting_file_count: int = store._get_query(table=File).count()

    # WHEN adding the modified bundle to the database
    new_bundle_obj, version = store.add_bundle(second_bundle_data)
    store.session.add(new_bundle_obj)
    store.session.add(version)
    store.session.commit()

    # THEN there should still be the same number of bundles
    assert store._get_query(table=Bundle).count() == starting_bundle_count
    # THEN there should be one more version
    assert store._get_query(table=Version).count() == starting_version_count + 1
    # THEN there should be two more files
    assert store._get_query(table=File).count() == starting_file_count + 2


def test_add_archive(archiving_task_id: int, populated_store: Store, spring_file_2: Path):
    """Test that adding an archive works as expected."""
    # GIVEN a file that is not archived
    non_archived_file: File = populated_store.get_files(file_path=spring_file_2.as_posix()).first()
    # WHEN adding an archive
    new_archive: Archive = populated_store.create_archive(
        file_id=non_archived_file.id, archiving_task_id=archiving_task_id
    )
    populated_store.session.add(new_archive)
    populated_store.session.commit()
    # THEN an archive should be created
    assert isinstance(new_archive, Archive)
    # THEN the archive should be reachable via the file
    assert non_archived_file.archive == new_archive


def test_add_archive_to_archived_file(
    archiving_task_id: int, populated_store: Store, spring_file_1: Path
):
    """Test that adding an archive to an already archived file raises an error."""
    # GIVEN a file that is archived
    non_archived_file: File = populated_store.get_files(file_path=spring_file_1.as_posix()).first()
    # WHEN adding an archive
    new_archive: Archive = populated_store.create_archive(
        file_id=non_archived_file.id, archiving_task_id=archiving_task_id
    )
    populated_store.session.add(new_archive)

    # THEN an SQLAlchemy error should be thrown
    with pytest.raises(IntegrityError):
        populated_store.session.commit()


def test_add_file(
    populated_store: Store,
    second_family_vcf: Path,
    family_tag_names: list[str],
    housekeeper_version_dir: Path,
    project_dir: Path,
):
    """Test to create a file with the add file method."""
    # GIVEN the path and the tags for a file

    # GIVEN a store populated with a bundle
    bundle: Bundle = populated_store.bundles().first()
    assert isinstance(bundle, Bundle)

    # WHEN using the add file method to create a new file object
    new_file: File = populated_store.add_file(
        file_path=second_family_vcf,
        bundle=bundle,
        tags=family_tag_names,
    )

    # THEN assert that the file is a file object
    assert isinstance(new_file, File)
    # THEN assert that the file is added to the latest version of the bundle
    assert new_file.version == bundle.versions[0]
    # THEN assert that the tags are added to the new file
    assert len(new_file.tags) == len(family_tag_names)
    for tag_obj in new_file.tags:
        assert isinstance(tag_obj, Tag)


def test_add_file_no_tags(
    populated_store: Store,
    second_family_vcf: Path,
    housekeeper_version_dir: Path,
    project_dir: Path,
):
    """Test to create a file with the add file method without tags."""
    # GIVEN a path for a file

    # GIVEN a store populated with a bundle
    bundle: Bundle = populated_store.bundles().first()
    assert isinstance(bundle, Bundle)

    # WHEN using the add file method to create a new file object
    new_file = populated_store.add_file(file_path=second_family_vcf, bundle=bundle)

    # THEN assert that the no tags where added to the file
    assert len(new_file.tags) == 0



"""This module handles adding things to the store"""

import datetime as dt
import logging
from pathlib import Path
from typing import Dict, Tuple

from sqlalchemy.orm import Session

from housekeeper.store.api.handlers.base import BaseHandler
from housekeeper.store.api.handlers.read import ReadHandler
from housekeeper.store.models import Archive, Bundle, File, Tag, Version

LOG = logging.getLogger(__name__)


class CreateHandler(BaseHandler):
    """Handles adding things to the store"""

    def __init__(self, session: Session):
        super().__init__(session=session)
        CreateHandler.get_bundle_by_name = ReadHandler.get_bundle_by_name
        CreateHandler.get_tag = ReadHandler.get_tag
        CreateHandler.get_version_by_date_and_bundle_name = (
            ReadHandler.get_version_by_date_and_bundle_name
        )

    def new_bundle(self, name: str, created_at: dt.datetime = None) -> Bundle:
        """Create a new file bundle."""
        new_bundle = Bundle(name=name, created_at=created_at)
        LOG.debug("Created new bundle: %s", new_bundle.name)
        return new_bundle

    def add_bundle(self, data: dict) -> Tuple[Bundle, Version] | None:
        """Build a new bundle version of files.

        The format of the input dict is defined in the `schema` module.
        """
        bundle_obj = self.get_bundle_by_name(bundle_name=data["name"])
        # These lines can be removed when decoupled from CG
        created_at = data.get("created_at", data.get("created"))
        expires_at = data.get("expires_at", data.get("expires"))
        if bundle_obj and self.get_version_by_date_and_bundle_name(
            version_date=created_at, bundle_name=bundle_obj.name
        ):
            LOG.debug("version of bundle already added")
            return None

        if bundle_obj is None:
            bundle_obj = self.new_bundle(name=data["name"], created_at=created_at)

        version_obj = self.new_version(created_at=created_at, expires_at=expires_at)
        self._add_files_to_version(data["files"], version_obj)

        version_obj.bundle = bundle_obj
        return bundle_obj, version_obj

    def _add_files_to_version(self, files: list[dict], version_obj: Version) -> None:
        """Create file objects and the tags and add them to a version object"""

        tag_names = set(tag_name for file_data in files for tag_name in file_data["tags"])
        tag_map = self._build_tags(tag_names)

        for file_data in files:
            # This if can be removed after decoupling
            if isinstance(file_data["path"], str):
                paths = [file_data["path"]]
            else:
                paths = file_data["path"]
            for path in paths:
                LOG.debug("adding file: %s", path)
                if not Path(path).exists():
                    raise FileNotFoundError(path)
                tags = [tag_map[tag_name] for tag_name in file_data["tags"]]
                new_file = self.new_file(path, to_archive=file_data["archive"], tags=tags)
                version_obj.files.append(new_file)

    def new_version(self, created_at: dt.datetime, expires_at: dt.datetime = None) -> Version:
        """Create a new bundle version."""
        LOG.debug("Created new version")
        new_version = Version(created_at=created_at, expires_at=expires_at)
        return new_version

    def add_version(
        self,
        data: dict,
        bundle: Bundle,
    ) -> Version:
        """Build a new version object and add it to an existing bundle"""
        created_at = data.get("created_at", data.get("created"))
        if self.get_version_by_date_and_bundle_name(
            version_date=created_at, bundle_name=bundle.name
        ):
            LOG.debug("version of bundle already added")
            return None

        version_obj = self.new_version(
            created_at=created_at,
            expires_at=data.get("expires_at", data.get("expires")),
        )
        if data.get("files"):
            self._add_files_to_version(data["files"], version_obj)

        version_obj.bundle = bundle
        return version_obj

    def add_file(
        self,
        file_path: Path,
        bundle: Bundle,
        to_archive: bool = False,
        tags: list[str] = None,
    ) -> File:
        """Build a new file object and add it to the latest version of an existing bundle."""
        version = bundle.versions[0]
        tags = tags or []
        tag_objs = [tag_obj for tag_name, tag_obj in self._build_tags(tags).items()]
        file_path_to_use: str = str(file_path)
        new_file = self.new_file(
            path=file_path_to_use,
            to_archive=to_archive,
            tags=tag_objs,
        )
        new_file.version = version
        return new_file

    def _build_tags(self, tag_names: list[str]) -> Dict[str, Tag]:
        """Build a list of tag objects.

        Take a list of tags, if a tag does not exist create a new tag object.
        Map the tag name to a tag object and return a list of those
        """
        tags = {}
        for tag_name in tag_names:
            tag_obj = self.get_tag(tag_name)
            if tag_obj is None:
                LOG.debug("create new tag: %s", tag_name)
                tag_obj = self.new_tag(tag_name)
            tags[tag_name] = tag_obj
        return tags

    def new_file(
        self,
        path: str,
        checksum: str = None,
        to_archive: bool = False,
        tags: list[Tag] = None,
    ) -> File:
        """Create a new file object based on the information given."""
        return File(path=path, checksum=checksum, to_archive=to_archive, tags=tags)

    def new_tag(self, name: str, category: str = None) -> Tag:
        """Create a new tag object based on the information given."""
        new_tag = Tag(name=name, category=category)
        return new_tag

    def create_archive(self, file_id: int, archiving_task_id: int) -> Archive:
        """Creates an archive object to the given file, with the given archive task id."""
        return Archive(file_id=file_id, archiving_task_id=archiving_task_id)



"""Tests for date parsing module"""
import datetime

import pytest

from housekeeper.date import get_date, match_date


def test_match_date_dash():
    """Test to match common string formated date"""
    # GIVEN a datestring separated by '-'
    date_str = "2015-05-10"
    # WHEN checking if it is a valid datestring
    res = match_date(date_str)
    # THEN assert the result is True
    assert res is True


def test_match_date_dot():
    """Test to match common string formated date"""
    # GIVEN a datestring separated by '.'
    date_str = "2015.05.10"
    # WHEN checking if it is a valid datestring
    res = match_date(date_str)
    # THEN assert the result is True
    assert res is True


def test_match_invalid_date():
    """Test to match a bad formated date string"""
    # GIVEN a datestring without separators
    date_str = "20150510"
    # WHEN checking if it is a valid datestring
    res = match_date(date_str)
    # THEN assert the result is False
    assert res is False


def test_match_non_date():
    """Test to match non date string"""
    # GIVEN a datestring not even similar to a date
    date_str = "hello"
    # WHEN checking if it is a valid datestring
    res = match_date(date_str)
    # THEN assert the result is False
    assert res is False


def test_valid_date_str():
    """Test to get a date object from a valid date string"""
    # GIVEN a datestring separated by '-'
    date_str = "2015-05-10"
    # WHEN converting to a datetime object
    date_obj = get_date(date_str)
    # THEN assert a succesfull conversion
    assert isinstance(date_obj, datetime.datetime)


def test_valid_date_str_no_value():
    """Test get a datetime object when date string has no value"""
    # GIVEN no datestring
    date_str = None
    # WHEN fetching a date object
    date_obj = get_date(date_str)
    # THEN assert a valid date was returned
    assert isinstance(date_obj, datetime.datetime)


def test_datetime_str():
    """Test get a datetime object the string is a datetime str"""
    # GIVEN a datestring directly from datetime
    date = datetime.datetime.now()
    date_str = str(date)

    # WHEN converting the string to a datetime object
    new_date = get_date(date_str)

    # THEN assert the new date is the same as original
    assert new_date == date


def test_datetime_timestamp(timestamp):
    """Test get a datetime object the string is a datetime str"""
    # GIVEN a datestring directly from datetime
    date_str = str(timestamp)

    # WHEN converting the string to a datetime object
    new_date = get_date(date_str)

    # THEN assert the new date is the same as original
    assert new_date == timestamp



"""Tests for the include module"""
import datetime
from pathlib import Path

import pytest

from housekeeper import include
from housekeeper.exc import VersionIncludedError
from housekeeper.store import models


def test_checksum(checksum: str, checksum_file: Path):
    """Test calculate checksum"""
    # GIVEN a file with a specific checksum

    # WHEN calculating the checksum
    calculated_checksum = include.checksum(checksum_file)

    # THEN it should match
    assert calculated_checksum == checksum


def test_include_version(project_dir: Path, version_obj: models.Version):
    """Test to include a version

    This should test if hardlinks where created when running the include function
    """
    global_root_path = project_dir
    version = version_obj
    # GIVEN a new directory as the global root
    file_names = [Path(file_obj.path).name for file_obj in version.files]
    date_str = str(version.created_at.date())
    assert len(list(global_root_path.iterdir())) == 0

    # WHEN including a bundle version
    include.include_version(str(global_root_path), version)

    # THEN it should link the files to the new destination
    version_root_path = global_root_path / version.bundle.name / date_str
    assert len(list(version_root_path.iterdir())) == len(version.files)
    for file_obj, file_name in zip(version.files, file_names):
        # ... and the file path should be replace with relative "internal" path
        assert (global_root_path / file_obj.path).exists()
        assert (global_root_path / file_obj.path).is_file()
        assert Path(file_obj.path) == Path(version.bundle.name) / date_str / file_name


def test_include_version_already_included(project_dir: Path, version_obj: models.Version):
    """Test to include files from version when they where already included"""
    # GIVEN a version which is already marked as included
    version_obj.included_at = datetime.datetime.now()

    # WHEN trying to "include it"
    with pytest.raises(VersionIncludedError):
        # THEN it should raise an exception
        include.include_version(project_dir, version_obj)



"""Tests for the get tags cli functionality."""

import logging

from click import Context
from click.testing import CliRunner, Result
from sqlalchemy.orm import Query

from housekeeper.cli.add import tag_cmd as add_tags
from housekeeper.cli.get import tag_cmd


def test_get_tags_empty(cli_runner: Context, base_context: CliRunner, caplog):
    """Test to get tags from an empty database"""
    caplog.set_level(logging.DEBUG)
    # GIVEN an empty database, log_output and a cli runner

    # WHEN getting all tags
    res: Result = cli_runner.invoke(tag_cmd, obj=base_context)

    # THEN assert it exits without problems
    assert res.exit_code == 0
    # THEN assert it communicates that the tags where not found
    assert "Could not find any of the specified tags" in caplog.text


def test_get_tags(cli_runner: Context, base_context: CliRunner, caplog):
    """Test to get tags from a database with some tags"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a database with some tags
    tag_names: list[str] = ["tag1", "tag2", "tag3"]
    cli_runner.invoke(add_tags, tag_names, obj=base_context)
    tag_query: Query = base_context["store"].get_tags()
    assert tag_query

    # WHEN getting all tags
    res: Result = cli_runner.invoke(tag_cmd, obj=base_context)

    # THEN assert it exits without problems
    assert res.exit_code == 0
    assert tag_names[0] in caplog.text


def test_get_tags_non_existing(
    cli_runner: Context, base_context: CliRunner, non_existent_tag_name, caplog
):
    """Test to get a non-existing tag from a database with some tags"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a database with some tags
    tag_names: list[str] = ["tag1", "tag2", "tag3"]
    cli_runner.invoke(add_tags, tag_names, obj=base_context)

    # WHEN getting tag that does not exist
    other_tag: str = non_existent_tag_name
    res: Result = cli_runner.invoke(tag_cmd, ["-n", other_tag], obj=base_context)

    # THEN assert it exits without problems
    assert res.exit_code == 0
    assert "Could not find any of the specified tags" in caplog.text


def test_get_tags_json(cli_runner: Context, base_context: CliRunner, caplog):
    """Test to get tags from a database with some tags in json format"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a database with some tags
    tag_names: list[str] = ["tag1", "tag2", "tag3"]
    cli_runner.invoke(add_tags, tag_names, obj=base_context)
    tag_query: Query = base_context["store"].get_tags()
    assert tag_query

    # WHEN getting all tags
    res: Result = cli_runner.invoke(tag_cmd, ["--json"], obj=base_context)

    # THEN assert it exits without problems
    assert res.exit_code == 0



"""This module defines a BaseHandler class holding different models"""

from typing import Type
from housekeeper.store.models import Archive, Bundle, File, Version, Tag, Model
from sqlalchemy.orm import Query, Session


class BaseHandler:
    """This is a base class holding different models."""

    def __init__(self, session: Session):
        self.session = session

    def _get_query(self, table: Type[Model]) -> Query:
        """Return a query for the given table."""
        return self.session.query(table)

    def _get_join_version_bundle_query(self) -> Query:
        """Return version bundle query."""
        return self._get_query(table=Version).join(Version.bundle)

    def _get_join_file_tag_query(self) -> Query:
        """Return file tag query."""
        return self._get_query(table=File).join(File.tags)

    def _get_join_version_query(self, query: Query):
        """Joins the given query with the version table."""
        return query.join(Version)

    def _get_join_bundle_to_file_query(self, file_query: Query) -> Query:
        """Joins bundle information to a File query."""
        return file_query.join(File.version).join(Version.bundle)

    def _get_join_file_tags_archive_query(self) -> Query:
        """Returns a File query joined with bundle, tags and archive information."""
        return self._get_join_bundle_to_file_query(
            file_query=self._get_join_file_tag_query()
        ).outerjoin(File.archive)



"""Tests for delete CLI functions"""
import logging

from click import Context
from click.testing import CliRunner

from housekeeper.cli import delete
from housekeeper.store.api.core import Store


def test_delete_non_existing_bundle(base_context: Context, cli_runner: CliRunner, caplog):
    """Test to delete a non existing bundle"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a store and a cli runner
    # GIVEN a case name that does not exist
    case_name = "hello"
    assert not base_context["store"].get_bundle_by_name(bundle_name=case_name)

    # WHEN trying to delete a non existing bundle
    result = cli_runner.invoke(delete.bundle_cmd, [case_name], obj=base_context)

    # THEN assert it exits non zero
    assert result.exit_code == 1
    # THEN it should communicate that the bundle was not found
    assert f"bundle {case_name} not found" in caplog.text


def test_delete_existing_bundle_with_version(
    populated_context: Context, cli_runner: CliRunner, case_id: str, caplog
):
    """Test to delete an existing bundle with versions"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a store and a cli runner
    store: Store = populated_context["store"]
    # GIVEN a bundle with versions
    bundle_obj = store.get_bundle_by_name(bundle_name=case_id)
    assert len(bundle_obj.versions) > 0

    # WHEN trying to delete a bundle
    result = cli_runner.invoke(delete.bundle_cmd, [case_id], obj=populated_context)

    # THEN it should ask if you are sure
    assert "Can not delete bundle, please remove all versions first" in caplog.text


def test_delete_existing_bundle_no_versions_no_confirmation(
    populated_context: Context, cli_runner: CliRunner, case_id: str, caplog
):
    """Test to delete an existing bundle without confirmation"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a store and a cli runner
    store: Store = populated_context["store"]
    # GIVEN a bundle without versions
    cli_runner.invoke(delete.version_cmd, ["-b", case_id], obj=populated_context, input="Yes")
    bundle_obj = store.get_bundle_by_name(bundle_name=case_id)
    assert len(bundle_obj.versions) == 0

    # WHEN trying to delete a bundle
    result = cli_runner.invoke(delete.bundle_cmd, [case_id], obj=populated_context, input="no")
    # THEN assert it exits non zero
    assert result.exit_code == 1


def test_delete_existing_bundle_no_versions_with_confirmation(
    populated_context: Context, cli_runner: CliRunner, case_id: str, caplog
):
    """Test to delete an existing bundle without confirmation"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a store and a cli runner
    store: Store = populated_context["store"]
    # GIVEN a bundle without versions
    cli_runner.invoke(delete.version_cmd, ["-b", case_id], obj=populated_context, input="Yes")
    bundle_obj = store.get_bundle_by_name(bundle_name=case_id)
    assert len(bundle_obj.versions) == 0

    # WHEN trying to delete a bundle
    result = cli_runner.invoke(delete.bundle_cmd, [case_id], obj=populated_context, input="Yes")
    # THEN assert it exits non zero
    assert result.exit_code == 0
    # THEN it should communicate that it was deleted
    assert f"Deleted bundle {case_id}" in caplog.text



"""Code for CLI get"""
import json as jsonlib
import logging

import click
from rich.console import Console

from housekeeper.store.api import schema
from housekeeper.store.api.core import Store
from housekeeper.store.models import Version

from .tables import (
    get_bundles_table,
    get_files_table,
    get_tags_table,
    get_versions_table,
)

LOG = logging.getLogger(__name__)


@click.group()
def get():
    """Get info from database"""


@get.command("bundle")
@click.argument("bundle-name", required=False)
@click.option("-i", "--bundle-id", type=int, help="Search for a bundle with bundle id")
@click.option("-j", "--json", is_flag=True, help="Output to json format")
@click.option("-v", "--verbose", is_flag=True, help="List files from latest version")
@click.option(
    "-c",
    "--compact",
    is_flag=True,
    help="print compact filenames IFF verobe flag present",
)
@click.pass_context
def bundle_cmd(context, bundle_name, bundle_id, json, verbose, compact):
    """Get bundle information from database"""
    store: Store = context.obj["store"]
    bundles = store.bundles()

    if bundle_name:
        bundle = store.get_bundle_by_name(bundle_name=bundle_name)
        bundles = [bundle] if bundle else []

    if bundle_id:
        bundle = store.get_bundle_by_id(bundle_id=bundle_id)
        bundles = [bundle] if bundle else []

    if not bundles:
        LOG.info("Could not find any bundles")
        return
    template = schema.BundleSchema()
    result = []
    for bundle in bundles:
        result.append(template.dump(bundle))

    if json:
        click.echo(jsonlib.dumps(result, indent=4, sort_keys=True))
        return
    console = Console()
    console.print(get_bundles_table(result))
    if verbose:
        for bundle in bundles:
            if len(bundle.versions) == 0:
                LOG.info("No versions found for bundle %s", bundle.name)
                return
            version_obj = bundle.versions[0]
            context.invoke(version_cmd, version_id=version_obj.id, verbose=True, compact=compact)


@get.command("version")
@click.option("-b", "--bundle-name", help="Fetch all versions from a bundle")
@click.option("-i", "--version-id", type=int, help="Fetch a specific version")
@click.option("-j", "--json", is_flag=True, help="Output to json format")
@click.option("-v", "--verbose", is_flag=True, help="print additional information")
@click.option(
    "-c",
    "--compact",
    is_flag=True,
    help="print compact filenames IFF verobe flag present",
)
@click.pass_context
def version_cmd(context, bundle_name, json, version_id, verbose, compact):
    """Get versions from database"""
    store: Store = context.obj["store"]
    if not (bundle_name or version_id):
        LOG.info("Please select a bundle or a version")
        return
    if bundle_name:
        bundle = store.get_bundle_by_name(bundle_name=bundle_name)
        if not bundle:
            LOG.info("Could not find bundle %s", bundle_name)
            return
        version_objs = bundle.versions

    if version_id:
        version: Version = store.get_version_by_id(version_id=version_id)
        if not version:
            LOG.warning("Could not find version %s", version_id)
            raise click.Abort
        version_objs = [version]

    version_template = schema.VersionSchema()
    result = []
    for version_obj in version_objs:
        bundle = store.get_bundle_by_id(bundle_id=version_obj.bundle_id)
        res = version_template.dump(version_obj)
        res["bundle_name"] = bundle.name
        result.append(res)

    if json:
        click.echo(jsonlib.dumps(result))
        return

    console = Console()
    console.print(get_versions_table(result))
    if not verbose:
        return

    for version_obj in version_objs:
        context.invoke(files_cmd, version_id=version_obj.id, verbose=True, compact=compact)


@get.command("file")
@click.option("-t", "--tag", "tag_names", multiple=True, help="filter by file tag")
@click.option("-v", "--version", "version_id", type=int, help="filter by version of the bundle")
@click.option("-V", "--verbose", is_flag=True, help="print additional information")
@click.option("-j", "--json", is_flag=True, help="Output to json format")
@click.option("-c", "--compact", is_flag=True, help="print compact filenames")
@click.argument("bundle", required=False)
@click.pass_context
def files_cmd(
    context,
    tag_names: list[str],
    version_id: int,
    verbose: bool,
    bundle: str,
    json: bool,
    compact: bool,
):
    """Get files from database"""
    store: Store = context.obj["store"]
    file_objs = store.get_files(bundle_name=bundle, tag_names=tag_names, version_id=version_id)
    template = schema.FileSchema()
    result = []
    for file in file_objs:
        result.append(template.dump(file))
    if json:
        click.echo(jsonlib.dumps(result))
        return
    console = Console()
    console.print(get_files_table(result, verbose=verbose, compact=compact))


@get.command("tag")
@click.option("-j", "--json", is_flag=True, help="Output to json format")
@click.option("-n", "--name", multiple=True, help="Specify a tag name")
@click.pass_context
def tag_cmd(context, json, name):
    """Get the tags from database"""
    store: Store = context.obj["store"]
    LOG.info("Fetch tags")
    tag_objs = store.get_tags()
    template = schema.TagSchema()
    result = []
    for tag_obj in tag_objs:
        if name and (tag_obj.name not in name):
            continue
        LOG.debug("Use tag %s", tag_obj.name)
        result.append(template.dump(tag_obj))
    if not result:
        LOG.info("Could not find any of the specified tags [%s]", ", ".join(name))
        return
    if json:
        click.echo(jsonlib.dumps(result))
        return
    console = Console()
    console.print(get_tags_table(result))



"""Tests for adding tags via CLI"""
import logging

from housekeeper.cli.add import tag_cmd
from housekeeper.store.api.core import Store
from housekeeper.store.models import File


def test_add_tags_no_args(populated_context, cli_runner, caplog):
    """Test to add a two tags to the database"""
    # GIVEN a context with a populated store, and a cli runner
    caplog.set_level(logging.DEBUG)
    # GIVEN that there is no tag input
    # WHEN trying to add two tags to the an existing file
    result = cli_runner.invoke(tag_cmd, [], obj=populated_context)
    # THEN assert it has a zero exit status
    assert result.exit_code == 1
    # THEN check that the correct information is logged
    assert "No tags provided" in caplog.text


def test_add_two_tags(populated_context, cli_runner, caplog):
    """Test to add a two tags to the database"""
    # GIVEN a context with a populated store, and a cli runner
    caplog.set_level(logging.DEBUG)
    # GIVEN two new tags
    tag1 = "new-tag"
    tag2 = "other-tag"

    # WHEN trying to add two tags to the an existing file
    result = cli_runner.invoke(tag_cmd, [tag1, tag2], obj=populated_context)

    # THEN assert it has a zero exit status
    assert result.exit_code == 0
    # THEN check that the tags are logged
    assert tag1 in caplog.text
    # THEN check that the tags are logged
    assert tag2 in caplog.text
    # THEN assert that the tags are added to the data base
    db_tags = set([tag.name for tag in populated_context["store"].get_tags()])
    assert db_tags.intersection(set([tag1, tag2]))


def test_add_existing_tag_existing_file(populated_context, cli_runner, caplog):
    """Test to add a existing tag to a file that exists"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a populated store, and a cli runner
    store: Store = populated_context["store"]
    # GIVEN a existing file id
    file_id = 1
    file_obj: File = store.get_file_by_id(file_id=file_id)
    # GIVEN that the new tag already exists for the file
    tag = file_obj.tags[0].name

    # WHEN trying to add the existing tag to the file
    result = cli_runner.invoke(tag_cmd, [tag, "-f", str(file_id)], obj=populated_context)
    # THEN assert it has a non zero exit status
    assert result.exit_code == 0
    # THEN check that it communicates that the tag existed
    assert f"{tag}: tag already added" in caplog.text


def test_add_tag_existing_file(populated_context, cli_runner, caplog):
    """Test to add a non existing tag to a file that exists"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a populated store, and a cli runner
    store: Store = populated_context["store"]
    # GIVEN a existing file id
    file_id = 1
    file_obj: File = store.get_file_by_id(file_id=file_id)
    assert file_obj
    # GIVEN a new tag
    tag = "new-tag"

    # WHEN trying to add a tag to the existing file
    result = cli_runner.invoke(tag_cmd, [tag, "-f", str(file_id)], obj=populated_context)
    # THEN assert it has a zero exit status
    assert result.exit_code == 0
    # THEN check that the tag is displayed in the output
    assert f"{tag}: tag created" in caplog.text


def test_add_tag_non_existing_file(populated_context, cli_runner, caplog):
    """Test to add a tag to a file that not exist"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a populated store and a cli runner
    store: Store = populated_context["store"]
    # GIVEN a non existing file id
    missing_file_id = 42
    file_obj = store.get_file_by_id(file_id=missing_file_id)
    assert not file_obj

    # WHEN trying to add a tag to the non existing file
    result = cli_runner.invoke(
        tag_cmd, ["new-tag", "-f", str(missing_file_id)], obj=populated_context
    )
    # THEN assert it has a non zero exit status
    assert result.exit_code == 1
    # THEN check that the error message is displayed
    assert "unable to find file" in caplog.text



"""Tests foe adding files via CLI"""
import logging
from pathlib import Path

from click import Context
from click.testing import CliRunner
from housekeeper.cli.add import file_cmd
from housekeeper.store import Store
from housekeeper.store.models import Bundle, Version

NEW_FILE_ADDED: str = "new file added"


def test_add_file_non_existing_bundle(
    base_context: Context,
    cli_runner: CliRunner,
    case_id: str,
    second_sample_vcf: Path,
    caplog,
):
    """Test to add a file to a non existing bundle"""
    caplog.set_level(logging.DEBUG)

    # GIVEN a context with a empty store and a cli runner
    unknown_bundle_name = case_id
    # WHEN trying to add a bundle
    result = cli_runner.invoke(
        file_cmd, [str(second_sample_vcf), "-b", unknown_bundle_name], obj=base_context
    )

    # THEN assert it fails
    assert result.exit_code == 1
    # THEN check that the proper information is displayed
    assert f"unknown bundle: {unknown_bundle_name}" in caplog.text


def test_add_file_json_and_file_path(
    base_context: Context,
    cli_runner: CliRunner,
    file_data_json: str,
    sample_vcf: Path,
    caplog,
):
    """Test to add a file to a non existing bundle"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a empty store and a cli runner

    # WHEN trying to add a bundle
    result = cli_runner.invoke(
        file_cmd, [str(sample_vcf), "--json", file_data_json], obj=base_context
    )

    # THEN assert it fails
    assert result.exit_code == 1
    # THEN check that the proper information is displayed
    assert "Can not input both json and path" in caplog.text


def test_add_file_no_input(base_context: Context, cli_runner: CliRunner, caplog):
    """Test to add a file without specifying any input"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a empty store and a cli runner

    # WHEN trying to add a bundle
    result = cli_runner.invoke(file_cmd, [], obj=base_context)

    # THEN assert it fails since there is no valid input
    assert result.exit_code == 1
    # THEN check that the proper information is displayed
    assert "Please input json or path" in caplog.text


def test_add_non_existing_file(base_context: Context, cli_runner: CliRunner, caplog):
    """Test to add a file without specifying any input"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a empty store and a cli runner
    non_existing = Path("hello.txt")
    assert not non_existing.exists()

    # WHEN trying to add a bundle
    result = cli_runner.invoke(file_cmd, [str(non_existing)], obj=base_context)

    # THEN assert it fails since there is no valid input
    assert result.exit_code == 1
    # THEN check that the proper information is displayed
    assert f"File: {non_existing} does not exist" in caplog.text


def test_add_file_existing_bundle_with_include(
    populated_context: dict,
    cli_runner: CliRunner,
    case_id: str,
    second_sample_vcf: Path,
    caplog,
    housekeeper_version_dir: Path,
    project_dir: Path,
):
    """Test to add a file to a existing bundle"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a populated store and a cli runner
    bundle_name = case_id

    store: Store = populated_context["store"]
    bundle_obj: Bundle = store.bundles().first()
    version_obj: Version = bundle_obj.versions[0]

    # WHEN trying to add the file to a bundle
    result = cli_runner.invoke(
        file_cmd,
        [str(second_sample_vcf), "-b", bundle_name],
        obj=populated_context,
    )

    # THEN assert it succedes
    assert result.exit_code == 0
    # THEN check that the proper information is displayed
    assert NEW_FILE_ADDED in caplog.text
    # THEN check that the file has been included in the version and that the relative path is given
    assert Path(housekeeper_version_dir, second_sample_vcf.name).exists()
    assert Path(housekeeper_version_dir, second_sample_vcf.name).is_file()
    assert Path(version_obj.files[2].path) == Path(
        version_obj.relative_root_dir, second_sample_vcf.name
    )


def test_add_file_existing_bundle_without_include(
    populated_context: Context,
    cli_runner: CliRunner,
    case_id: str,
    second_sample_vcf: Path,
    caplog,
    housekeeper_version_dir: Path,
):
    """Test to add a file to a existing bundle"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a populated store and a cli runner
    bundle_name = case_id

    # WHEN trying to add the file to a bundle
    result = cli_runner.invoke(
        file_cmd,
        [str(second_sample_vcf), "-b", bundle_name, "-kip"],
        obj=populated_context,
    )

    # THEN assert the program exits with success
    assert result.exit_code == 0
    # THEN check that the proper information is displayed
    assert NEW_FILE_ADDED in caplog.text


def test_add_file_json(
    populated_context: Context,
    cli_runner: CliRunner,
    file_data_json: str,
    housekeeper_version_dir: Path,
    caplog,
):
    """Test to add a file in json format to a non existing bundle"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a populated store and a cli runner

    # WHEN trying to add a bundle
    result = cli_runner.invoke(file_cmd, ["--json", file_data_json], obj=populated_context)

    # THEN assert it fails
    assert result.exit_code == 0
    # THEN check that the proper information is displayed
    assert NEW_FILE_ADDED in caplog.text



import pytest
from sqlalchemy.orm import Query

from housekeeper.store import Store
from housekeeper.store.filters.file_filters import (
    filter_files_by_id,
    filter_files_by_is_archived,
    filter_files_by_path,
    filter_files_by_tags,
)
from housekeeper.store.models import File


def test_filter_files_by_id_returns_query(populated_store: Store):
    """Test that filter_files_by_id returns a Query object."""

    # GIVEN a store with a file
    file: File = populated_store._get_query(table=File).first()
    assert file

    # WHEN retrieving the file by id
    file_query: Query = filter_files_by_id(
        files=populated_store._get_query(table=File), file_id=file.id
    )

    # THEN a query should be returned
    assert isinstance(file_query, Query)


def test_filter_files_by_id_returns_the_correct_file(populated_store: Store):
    """Test filtering files by id returns the correct file."""

    # GIVEN a store with a file
    file: File = populated_store._get_query(table=File).first()
    assert file

    # WHEN retrieving the file by id
    file_query: Query = filter_files_by_id(
        files=populated_store._get_query(table=File),
        file_id=file.id,
    )

    # THEN a file should be returned
    filtered_file: File = file_query.first()
    assert isinstance(filtered_file, File)

    # THEN the id should match
    assert filtered_file.id == file.id


def test_filter_files_by_path_returns_the_correct_file(populated_store: Store):
    """ "Test filtering files by path returns the correct file."""
    # GIVEN a store with some files
    files = populated_store._get_query(table=File)

    file: File = files.first()
    assert file

    # WHEN retrieving the file by path
    file_query: Query = filter_files_by_path(
        files=files,
        file_path=file.path,
    )

    # THEN a file should be returned
    filtered_file: File = file_query.first()
    assert isinstance(filtered_file, File)

    # THEN the path should match
    assert filtered_file.path == file.path


@pytest.mark.parametrize("file_id", [-1, None, "non-existent"])
def test_filter_files_by_id_returns_none_for_invalid_id(populated_store: Store, file_id):
    """Test that filter_files_by_id returns None for invalid file ids."""

    # WHEN retrieving the file by an invalid id
    file_query: Query = filter_files_by_id(
        files=populated_store._get_query(table=File), file_id=file_id
    )

    # THEN no file should be returned
    filtered_file: File = file_query.first()
    assert filtered_file is None


def test_filter_files_by_tags_returns_correct_files(populated_store: Store):
    """Test filtering files by tags."""

    # GIVEN a store with files
    file: File = populated_store._get_query(table=File).first()
    tag_names: list[str] = [tag.name for tag in file.tags]

    # WHEN filtering files by tags
    filtered_files_query = filter_files_by_tags(
        files=populated_store._get_join_file_tag_query(),
        tag_names=tag_names,
    )

    filtered_files: list[File] = filtered_files_query.all()

    # THEN each file should have all the requested tags
    for filtered_file in filtered_files:
        assert all(tag.name in tag_names for tag in filtered_file.tags)


def test_filter_files_by_tags_returns_empty_list_when_no_files_match_tags(
    populated_store: Store,
):
    """Test filtering files by tags when no files match the given tags."""

    # GIVEN a store with files
    # GIVEN a tag that does not exist in any files in the store
    tag_name = "non_existent_tag"

    # WHEN filtering files by the non_existent tag
    filtered_files_query = filter_files_by_tags(
        files=populated_store._get_join_file_tag_query(),
        tag_names=[tag_name],
    )

    # THEN the filtered files list should be empty
    assert len(filtered_files_query.all()) == 0


def test_filter_files_by_archive_true(populated_store: Store):
    """Tests the filtering for archived files."""

    # GIVEN as store with files

    # WHEN filtering by archived files
    archived_files_query: Query = filter_files_by_is_archived(
        files=populated_store._get_join_file_tags_archive_query(),
        is_archived=True,
    )

    # THEN all files returned should have an archive object linked to it
    for file in archived_files_query:
        assert file.archive


def test_filter_files_by_archive_false(populated_store: Store):
    """Tests the filtering for non-archived files."""

    # GIVEN as store with files

    # WHEN filtering on non-archived files
    non_archived_files_query: Query = filter_files_by_is_archived(
        files=populated_store._get_join_file_tags_archive_query(),
        is_archived=False,
    )

    # THEN none of the files returned should have an archive object linked to it
    for file in non_archived_files_query:
        assert file.archive is None



"""
This module handles finding things in the store/database
"""
import datetime as dt
import logging
from pathlib import Path

from sqlalchemy.orm import Query, Session

from housekeeper.store.api.handlers.base import BaseHandler
from housekeeper.store.filters.archive_filters import (
    ArchiveFilter,
    apply_archive_filter,
)
from housekeeper.store.filters.bundle_filters import BundleFilters, apply_bundle_filter
from housekeeper.store.filters.file_filters import FileFilter, apply_file_filter
from housekeeper.store.filters.tag_filters import TagFilter, apply_tag_filter
from housekeeper.store.filters.version_bundle_filters import (
    VersionBundleFilters,
    apply_version_bundle_filter,
)
from housekeeper.store.filters.version_filters import (
    VersionFilter,
    apply_version_filter,
)
from housekeeper.store.models import Archive, Bundle, File, Tag, Version

LOG = logging.getLogger(__name__)


class ReadHandler(BaseHandler):
    """Handler for searching the database"""

    def __init__(self, session: Session):
        super().__init__(session=session)

    def bundles(self):
        """Fetch bundles."""
        LOG.debug("Fetching all bundles")
        return self._get_query(table=Bundle)

    def get_bundle_by_id(self, bundle_id: int) -> Bundle:
        """Fetch a bundle by id from the store."""
        LOG.debug(f"Fetching bundle with id: {bundle_id}")
        return apply_bundle_filter(
            bundles=self._get_query(table=Bundle),
            filter_functions=[BundleFilters.FILTER_BY_ID],
            bundle_id=bundle_id,
        ).first()

    def get_bundle_by_name(self, bundle_name: str) -> Bundle:
        """Get a bundle by name from the store."""
        LOG.debug(f"Fetching bundle with name: {bundle_name}")
        return apply_bundle_filter(
            bundles=self._get_query(table=Bundle),
            filter_functions=[BundleFilters.FILTER_BY_NAME],
            bundle_name=bundle_name,
        ).first()

    def get_version_by_date_and_bundle_name(
        self, version_date: dt.datetime, bundle_name: str
    ) -> Version:
        return apply_version_bundle_filter(
            version_bundles=self._get_join_version_bundle_query(),
            filter_functions=[VersionBundleFilters.FILTER_BY_DATE_AND_NAME],
            version_date=version_date,
            bundle_name=bundle_name,
        ).first()

    def get_version_by_id(self, version_id: int) -> Version:
        """Fetch a version from the store."""
        LOG.debug(f"Fetching version with id: {version_id}")
        return apply_version_filter(
            versions=self._get_query(table=Version),
            filter_functions=[VersionFilter.FILTER_BY_ID],
            version_id=version_id,
        ).first()

    def get_tag(self, tag_name: str = None) -> Tag:
        """Return a tag from the database."""
        LOG.debug(f"Fetching tag with name: {tag_name}")
        return apply_tag_filter(
            tags=self._get_query(table=Tag),
            filter_functions=[TagFilter.FILTER_BY_NAME],
            tag_name=tag_name,
        ).first()

    def get_tags(self) -> Query:
        """Return all tags from the database."""
        LOG.debug("Fetching all tags")
        return self._get_query(table=Tag)

    def get_file_by_id(self, file_id: int) -> File | None:
        """Get a file by record id."""
        return apply_file_filter(
            files=self._get_query(table=File),
            filter_functions=[FileFilter.FILTER_BY_ID],
            file_id=file_id,
        ).first()

    def get_files(
        self,
        bundle_name: str = None,
        tag_names: list[str] = None,
        version_id: int = None,
        file_path: str = None,
    ) -> Query:
        """Fetches files from the store based on the specified filters.
        Args:
            bundle_name (str, optional): Name of the bundle to fetch files from.
            tag_names (list[str], optional): list of tags to filter files by.
            version_id (int, optional): ID of the version to fetch files from.
            path (str, optional): Path to the file to fetch.

        Returns:
            Query: A query that match the specified filters.
        """
        query = self._get_query(table=File)
        if bundle_name:
            LOG.debug(f"Fetching files from bundle {bundle_name}")
            query = apply_bundle_filter(
                bundles=query.join(File.version).join(Version.bundle),
                filter_functions=[BundleFilters.FILTER_BY_NAME],
                bundle_name=bundle_name,
            )

        if tag_names:
            formatted_tags = ",".join(tag_names)
            LOG.debug(f"Fetching files with tags in [{formatted_tags}]")

            query = apply_file_filter(
                files=query.join(File.tags),
                filter_functions=[FileFilter.FILTER_FILES_BY_TAGS],
                tag_names=tag_names,
            )

        if version_id:
            LOG.debug(f"Fetching files from version {version_id}")
            query = apply_version_filter(
                versions=query.join(File.version),
                filter_functions=[VersionFilter.FILTER_BY_ID],
                version_id=version_id,
            )

        if file_path:
            LOG.debug(f"Fetching file with path {file_path}")
            query = apply_file_filter(
                files=query,
                filter_functions=[FileFilter.FILTER_BY_PATH],
                file_path=file_path,
            )

        return query

    def get_files_before(
        self,
        bundle_name: str = None,
        tag_names: list[str] = None,
        before_date: dt.datetime = None,
    ) -> list[File]:
        """Return files before a specific date from store."""
        query = self.get_files(tag_names=tag_names, bundle_name=bundle_name)
        if before_date:
            query = apply_version_filter(
                versions=self._get_join_version_query(query),
                filter_functions=[VersionFilter.FILTER_BY_DATE],
                before_date=before_date,
            )
        return query.all()

    @staticmethod
    def get_files_not_on_disk(files: list[File]) -> list[File]:
        """Return list of files that are not on disk."""
        if not files:
            return []

        files_not_on_disk = [f for f in files if not Path(f.full_path).is_file()]
        return files_not_on_disk

    def get_archived_files_for_bundle(self, bundle_name: str, tags: list | None) -> list[File]:
        """Returns all files in the given bundle, with the given tags, and are archived."""
        files_filtered_on_bundle: Query = apply_bundle_filter(
            bundles=self._get_join_file_tags_archive_query(),
            bundle_name=bundle_name,
            filter_functions=[BundleFilters.FILTER_BY_NAME],
        )
        return apply_file_filter(
            files=files_filtered_on_bundle,
            filter_functions=[
                FileFilter.FILTER_FILES_BY_TAGS,
                FileFilter.FILTER_FILES_BY_IS_ARCHIVED,
            ],
            is_archived=True,
            tag_names=tags,
        ).all()

    def get_non_archived_files_for_bundle(self, bundle_name: str, tags: list | None) -> list[File]:
        """Returns all files in the given bundle, with the given tags, and are not archived."""
        files_filtered_on_bundle: Query = apply_bundle_filter(
            bundles=self._get_join_file_tags_archive_query(),
            bundle_name=bundle_name,
            filter_functions=[BundleFilters.FILTER_BY_NAME],
        )
        return apply_file_filter(
            files=files_filtered_on_bundle,
            filter_functions=[
                FileFilter.FILTER_FILES_BY_TAGS,
                FileFilter.FILTER_FILES_BY_IS_ARCHIVED,
            ],
            is_archived=False,
            tag_names=tags,
        ).all()

    def get_ongoing_archivals(self) -> list[Archive]:
        """Returns all archiving tasks in the archive table, for entries where the archiving
        field is empty."""
        return apply_archive_filter(
            archives=self._get_query(table=Archive),
            filter_functions=[ArchiveFilter.FILTER_ARCHIVING_ONGOING],
        ).all()

    def get_ongoing_retrievals(self) -> list[Archive]:
        """Returns all retrieval tasks in the archive table, for entries where the retrieved_at
        field is empty."""
        return apply_archive_filter(
            archives=self._get_query(table=Archive),
            filter_functions=[ArchiveFilter.FILTER_RETRIEVAL_ONGOING],
        ).all()

    def get_bundle_name_from_file_path(self, file_path: str) -> str:
        """Return the bundle name for the specified file."""
        return self.get_files(file_path=file_path).first().version.bundle.name

    def get_non_archived_files(self, tag_names: list[str], limit: int | None = None) -> list[File]:
        """Return all spring files which are not marked as archived in Housekeeper."""
        return (
            apply_file_filter(
                self._get_join_file_tags_archive_query(),
                filter_functions=[
                    FileFilter.FILTER_FILES_BY_TAGS,
                    FileFilter.FILTER_FILES_BY_IS_ARCHIVED,
                ],
                tag_names=tag_names,
                is_archived=False,
            )
            .limit(limit)
            .all()
        )

    def get_archives(
        self, archival_task_id: int = None, retrieval_task_id: int = None
    ) -> list[File] | None:
        """Returns all entries in the archive table with the specified archival/retrieval task id."""
        if not archival_task_id and not retrieval_task_id:
            return self._get_query(table=Archive).all()
        if archival_task_id and retrieval_task_id:
            return apply_archive_filter(
                archives=apply_archive_filter(
                    archives=self._get_query(table=Archive),
                    filter_functions=[ArchiveFilter.FILTER_BY_ARCHIVING_TASK_ID],
                    task_id=archival_task_id,
                ),
                filter_functions=[ArchiveFilter.FILTER_BY_RETRIEVAL_TASK_ID],
                task_id=retrieval_task_id,
            ).all()
        if archival_task_id:
            return apply_archive_filter(
                archives=self._get_query(table=Archive),
                filter_functions=[ArchiveFilter.FILTER_BY_ARCHIVING_TASK_ID],
                task_id=archival_task_id,
            ).all()

        return apply_archive_filter(
            archives=self._get_query(table=Archive),
            filter_functions=[ArchiveFilter.FILTER_BY_RETRIEVAL_TASK_ID],
            task_id=retrieval_task_id,
        ).all()



"""Code for building tables that are displayed in the terminal"""
import re
from pathlib import Path

from rich.table import Table


def get_tags_table(rows: list[dict]) -> Table:
    """Return a tag table"""
    table = Table(show_header=True, header_style="bold magenta")
    table.title = "[not italic]:bookmark:[/] Tags table [not italic]:bookmark:[/]"
    table.add_column("ID")
    table.add_column("Name")
    table.add_column("Category")
    table.add_column("Created")
    for tag_obj in rows:
        table.add_row(
            str(tag_obj["id"]),
            tag_obj["name"],
            tag_obj["category"],
            tag_obj["created_at"].split("T")[0],
        )
    return table


def get_files_table(rows: list[dict], verbose=False, compact=False) -> Table:
    """Return a tag table"""
    table = Table(show_header=True, header_style="bold magenta")
    table.title = "[not italic]:scroll:[/] Files table [not italic]:scroll:[/]"
    table.add_column("ID")
    table.add_column("File name")
    table.add_column("Tags")
    if compact:
        rows = squash_names(rows)
    for i, file_obj in enumerate(rows, 1):
        file_tags = ", ".join(tag["name"] for tag in file_obj["tags"])
        file_path = Path(file_obj["path"])
        file_name = file_path.name
        if verbose:
            file_name = file_obj["full_path"]

        if i % 2 == 0:
            table.add_row(str(file_obj["id"]), f"[yellow]{file_name}[/yellow]", file_tags)
        else:
            table.add_row(str(file_obj["id"]), f"[blue]{file_name}[/blue]", file_tags)
    return table


def get_bundles_table(rows: list[dict]) -> Table:
    """Return a bundles table"""
    table = Table(show_header=True, header_style="bold magenta")
    table.title = "[not italic]:package:[/] Bundle table [not italic]:package:[/]"
    table.add_column("ID")
    table.add_column("Bundle name")
    table.add_column("Version IDs")
    table.add_column("Created")
    for bundle_obj in rows:
        bundle_versions = ", ".join(str(version["id"]) for version in bundle_obj["versions"])
        table.add_row(
            str(bundle_obj["id"]),
            bundle_obj["name"],
            bundle_versions,
            bundle_obj["created_at"].split("T")[0],
        )
    return table


def get_versions_table(rows: list[dict]) -> Table:
    """Return a versions table"""
    table = Table(show_header=True, header_style="bold magenta")
    table.title = "[not italic]:closed_book:[/] Version table [not italic]:closed_book:[/]"
    table.add_column("ID")
    table.add_column("Bundle name")
    table.add_column("Nr files")
    table.add_column("Included")
    table.add_column("Archived")
    table.add_column("Created")
    table.add_column("Expires")
    for version_obj in rows:
        table.add_row(
            str(version_obj["id"]),
            version_obj["bundle_name"],
            str(len(version_obj["files"])),
            version_obj["included_at"].split("T")[0] if version_obj["included_at"] else "",
            version_obj["archived_at"].split("T")[0] if version_obj["archived_at"] else "",
            version_obj["created_at"].split("T")[0] if version_obj["created_at"] else "",
            version_obj["expires_at"].split("T")[0] if version_obj["expires_at"] else "",
        )
    return table


def squash_names(list_of_files: list[dict]) -> list[dict]:
    """If subsequent elements (filenames) in 'list_of_files' end in an integer- And that integer is
    following the previous- those are squashed when displayed.
    Example:

        ["asdf1.txt", "asdf2.txt"] becomes asdf[1-2].txt'

    Algorithm summary

        Traverse a list of dictionaries, where each dictionary represents
        a file with 'name', 'path' and 'tag', etc. For each name in the list ending
        with integer n, for every subsequent name ending in n+1, n+2, ... n+i displayed
        name wlll be: name[n-i].

        Input list will not be sorted.

        For each element e in the list, e will be searched by regular expression for integers
        at specific location, this is cached for next iteration where a comparision is done. If
        subsequent cache is updated, if not the result is printed. Print will look in the cache
        to check if current iteration has a subsequent integer ending. Tags associated to squashed
        filenames will be sorted and duplicates removed.
    """
    list_of_squashed = []
    tag_list = []
    if list_of_files == []:
        return list_of_squashed
    head = list_of_files[0]
    tail = list_of_files[1:]
    (previous_file, previous_counter, previous_suffix) = _get_suffix(head["path"])
    squash = [previous_counter]
    previous_hkjson = head
    for hk_json in tail + [{"path": ""}]:  # Add padding for final iteration
        (filename, counter, suffix) = _get_suffix(hk_json["path"])
        if counter == str(_to_int(previous_counter) + 1) and (previous_file == filename):
            squash = squash + [counter]
            tag_list = tag_list + (hk_json["tags"])
        else:
            if len(squash) == 1:  # only previous element in list
                list_of_squashed.append(previous_hkjson)
                squash = [counter]
            else:
                squashed_path = (
                    previous_file + "[" + squash[0] + "-" + squash[-1] + "]" + previous_suffix
                )
                previous_hkjson["path"] = squashed_path
                previous_hkjson["full_path"] = squashed_path
                previous_hkjson["tags"] = remove_duplicates(sorted(tag_list, key=lambda i: i["id"]))
                previous_hkjson["id"] = "-"
                list_of_squashed.append(previous_hkjson)
                squash = [counter]
                tag_list = []
        previous_counter = counter
        previous_suffix = suffix
        previous_file = filename
        previous_hkjson = hk_json
    return list_of_squashed


def remove_duplicates(tag_list: list[dict]) -> list[dict]:
    """Remove duoplicate elements from `tag_list`"""
    no_duplicates = []
    for i in tag_list:
        if i not in no_duplicates:
            no_duplicates.append(i)
    return no_duplicates


def _get_suffix(filename):
    """Split a filename if ending with an integer before suffix"""
    # re.split('(\d+)\.\w{3}$', "asdf1.asd")
    parsed = re.split("(\d+)\.(\w{2,3}$)", filename)
    if len(parsed) == 4:
        return (parsed[0], parsed[1], "." + parsed[2])
    return (filename, "", "")


def _to_int(string):
    """Cast to int, empty string becomes 0"""
    if string == "":
        return 0
    return int(string)
